/******************************************************************************
*
* project name:   FAT-Engine
* file name:      fatd10.c
* initial date:   29/04/2002
* author:         thomas.nussbaumer@gmx.net
*
* description:    FAT-Engine - Demo 10 (Door Demo)
*
* [+]/[-] opens a door
*
* $Id: fatd10.c,v 1.5 2002/08/01 12:40:28 tnussb Exp $
*
******************************************************************************/

//#define FAT_USEDEBUG

#include "../includes/fat.h"


#define  MAP_WIDTH        16
#define  MAP_HEIGHT       16

//-------------------------------------------------------------------
// defines the used map: 0 = no wall, 1 = texture1, 2 = texture2 etc.
// NOTE: the array tex64 which gets handed over to the engine as
//       part of the FATCONFIG structure has to as much textures as
//       you have specified in this map.
//       For example: if the highest number in the map is 3 then
//       tex64 has to hold 3 textures (64x64 pixels with 2 planes)
//-------------------------------------------------------------------

char map[MAP_WIDTH*MAP_HEIGHT] = {
     2, 1,   2, 2,   1, 2, 2, 2, 2, 2,   2,   1, 2,   2, 1, 2,
     2, 0,   0, 0,   0, 0, 0, 0, 0, 0,   0,   0, 0,   1, 0, 1,
     2, 0,   2, 2, 127, 2, 2, 0, 0, 2,   2,   0, 0,   0, 0, 2,
     2, 0,   2, 0,   0, 0, 2, 0, 0, 2,   0,   0, 0,   1, 1, 2,
     2, 0, 127, 0,   0, 0, 0, 0, 0, 0,   0,   0, 0,   0, 0, 2,
     2, 0,   2, 0,   0, 0, 2, 0, 0, 2,   0,   0, 0,   1, 0, 2,
     2, 0,   0, 2,   2, 2, 0, 0, 0, 0,   2,   2, 2,   0, 0, 2,
     2, 0,   0, 0,   0, 0, 0, 0, 0, 0,   0,   0, 0,   0, 0, 2,
     2, 0,   2, 2,   2, 2, 2, 0, 0, 2, 127,   1, 2,   2, 0, 2,
     2, 0,   0, 0,   0, 0, 2, 0, 0, 2,   0,   0, 0,   1, 0, 2,
     2, 0,   0, 0,   0, 0, 0, 0, 0, 0,   0,   0, 0,   2, 0, 2,
     2, 0,   0, 0,   0, 0, 0, 0, 0, 0,   0,   0, 0, 127, 0, 2,
     2, 1,   0, 1,   0, 0, 2, 0, 0, 2,   0,   0, 0,   1, 0, 2,
     2, 0,   0, 1,   2, 2, 2, 0, 0, 2,   2,   0, 2,   2, 0, 2,
     2, 0,   0, 1,   0, 0, 0, 0, 0, 0,   0,   0, 0,   0, 0, 1,
     2, 1,   2, 2,   1, 2, 2, 2, 2, 2,   2,   1, 2,   2, 1, 2
};


//-----------------------------------------------------------------------------
// include the utility routines ...
//-----------------------------------------------------------------------------
#include "utils.c"

//-----------------------------------------------------------------------------
// include file containing the texture and genericdata indices into fatdtx
// (automatically generated by TexMaker)
//-----------------------------------------------------------------------------
#include "indices.h"


short display_offsetx = 0; // will be set depending on calctype (TI89 or TI92p)
short display_offsety = 0; // will be set depending on calctype (TI89 or TI92p)


//=============================================================================
// initializes GUI (darkgray background, tiny logo and counter)
//=============================================================================
void InitGUI(COUNTER *counter) {
    static unsigned long empty[] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

    memset(GetPlane(LIGHT_PLANE),0,LCD_SIZE);
    memset(GetPlane(DARK_PLANE),0xff,LCD_SIZE);

    Sprite32(158-32+display_offsetx,display_offsety+2,16,empty,GetPlane(0),SPRT_AND);
    Sprite32(158-32+display_offsetx,display_offsety+2,16,FAT_GetTinyLogo(),GetPlane(0),SPRT_OR);
    Sprite32(158-32+display_offsetx,display_offsety+2,16,empty,GetPlane(1),SPRT_AND);
    Sprite32(158-32+display_offsetx,display_offsety+2,16,FAT_GetTinyLogo()+16,GetPlane(1),SPRT_OR);

    InitCounter(counter,158-6*5+display_offsetx,99-10+display_offsety,5,0);
    DrawCounter(counter);
}


//=============================================================================
// output frame count ever 10th frame
//=============================================================================
void frame_drawn(FATCONFIG* fc,COUNTER* counter) {
    short i;
    if (fc->frame_counter%10==0) {
        // we starting incrementing at 2nd position and not on first!
        for (i=counter->length-2;i>=0;i--) {

            counter->newcount[i] = counter->newcount[i]+1;
            if (counter->newcount[i] < 10) break;
            else                           counter->newcount[i]=0;
        }
        DrawCounter(counter);
    }
}


//=============================================================================
// add all doors defined in map
//=============================================================================
void CreateDoors(TEXCONFIG* tex_door, TEXCONFIG* tex_side) {
    short x, y;
    for (y=1; y<(MAP_HEIGHT-1); ++y) {
        for (x=1; x<(MAP_WIDTH-1); ++x) {
            if (map[y*MAP_WIDTH+x] == 127) {
                if (map[y*MAP_WIDTH+(x-1)]>0 && map[y*MAP_WIDTH+(x+1)]>0) {
                    FAT_AddDoor(x, y, DOOR_HORIZONTAL, DOOR_CLOSED, 0, tex_door, tex_side, NULL);
                }
                else {
                    FAT_AddDoor(x, y, DOOR_VERTICAL, DOOR_CLOSED, 0, tex_door, tex_side, NULL);
                }
            }
        }
    }

    FAT_AddDoor(2, 12, PWALL, DOOR_CLOSED, 0, tex_door, tex_side, NULL);
    map[12*MAP_WIDTH+2] = 127;

    FAT_AddDoor(13, 2, PWALL, DOOR_CLOSED, 0, tex_door, tex_side, NULL);
    map[2*MAP_WIDTH+13] = 127;


#if defined(FAT_USEDEBUG)
    y=0;
    FATDOOR* doors     = FAT_GetDoorsArray();
    short    doorcount = FAT_GetNrDoors();
    for (x=0,y=0;x<doorcount;x++,y+=5) {
        FAT_DebugUseSlot(y+0,"d.x",&(doors[x].xtile),FAT_DBG_SHORT_DEC);
        FAT_DebugUseSlot(y+1,"d.y",&(doors[x].ytile),FAT_DBG_SHORT_DEC);
        FAT_DebugUseSlot(y+2,"d.orient'n",&(doors[x].type),FAT_DBG_UCHAR_DEC);
        FAT_DebugUseSlot(y+3,"d.state",&(doors[x].state),FAT_DBG_UCHAR_DEC);
        FAT_DebugUseSlot(y+4,"d.position",&(doors[x].position),FAT_DBG_SHORT_DEC);
    }
#endif
}



//=============================================================================
// updates the door in front of the player
//=============================================================================
void UpdateDoor(FATCONFIG* fc, short open) {
    short tile_x = ((FAT_Cos16384(fc->cam_orientation) >> 8)+fc->cam_xpos)>>6;
    short tile_y = ((FAT_Sin16384(fc->cam_orientation) >> 8)+fc->cam_ypos)>>6;

    if (fc->cam_xpos == tile_x || fc->cam_ypos == tile_y) return; // cannot be on same tile!

    FATDOOR* door = FAT_GetDoor(tile_x,tile_y);

    if (!door) return;

    if (open) {
        if (door->state != DOOR_OPEN) {
            door->position+=2;
            if (door->position == 64) {
                door->state = DOOR_OPEN;
                *(fc->map_data + fc->map_width * tile_y + tile_x) = -127;
            }
            else {
                door->state = DOOR_OPENING;
            }
        }
    }
    else {
        if (door->state != DOOR_CLOSED) {
            door->position-=2;
            if (!door->position) {
                door->state = DOOR_CLOSED;
            }
            else {
                door->state = DOOR_CLOSING;
            }
            *(fc->map_data + fc->map_width * tile_y + tile_x) = 127;
        }
    }
}



//=============================================================================
// as usual ... our main function
//=============================================================================
void _main(void) {
    PADSTATES       key    = 0;
    unsigned char*  bg_img = NULL;
    short           i,j;
    unsigned short* dest;
    TEXCONFIG       textures[4];
    LCD_BUFFER      lcd;
    HANDLE          texfile = 0;
    unsigned char*  src     = 0;
    short           result;
    COUNTER         frames_display;

    FATCONFIG fc = {
        MAP_WIDTH,   // width of map
        MAP_HEIGHT,  // height of map
        map,         // pointer to map data
        textures,    // pointer to textures array
        0,           // number of sprites
        NULL,        // the sprite array
        96+3*64,     // start pos_x
        96,          // start pos_y
        ANGLE_SOUTH, // orientation
        NULL,        // destination plane0
        NULL,        // destination plane1
        NULL,        // background image
        0            // frame counter
    };

    //-------------------------------------------------------------------------
    // before we try to do anything else we have to initialize the engine
    // its only valid to proceed if FAT_LOADOKAY gets returned
    //-------------------------------------------------------------------------
    if ((result = FAT_LoadEngine()) != FAT_LOADOKAY) {
        switch(result) {
            case FAT_NOTGHOSTSPACE: ST_helpMsg("ERROR: client no exepacked!");break;
            case FAT_NOTFOUND:      ST_helpMsg("ERROR: no matching engine found");break;
            case FAT_LOCKFAILED:    ST_helpMsg("ERROR: locking engine failed");break;
            case FAT_OUTOFMEM:      ST_helpMsg("ERROR: out of memory");break;
            default:                ST_helpMsg("ERROR: unknown errorcode");break;
        }
        return;
    }

    //-------------------------------------------------------------------------
    // try to load the used textures from fatdtx ...
    //-------------------------------------------------------------------------
    if ((texfile = FAT_GetHandleOfFile("fatdtx"))) {
        if ((src = FAT_LockHandleOfFile(texfile))) {
            // load house wall textures
            if (!FAT_LoadTextures(textures,src+2,TEXIDX_HOUSE1,2)) {
                src = 0; // just to indicate something went wrong
            }
            // load door texture
            if (!FAT_LoadTextures(&textures[2],src+2,TEXIDX_DEMO3WALL,2)) {
                src = 0; // just to indicate something went wrong
            }
        }
    }

    //-------------------------------------------------------------------------
    // check if loading failed ...
    //-------------------------------------------------------------------------
    if (!texfile || !src) {
        FAT_UnloadEngine();
        if (texfile) FAT_UnlockHandleOfFile(texfile);
        ST_helpMsg("ERROR: cannot textures from fatdtx");
        return;
    }

    //-------------------------------------------------------------------------
    // add doors defined in map
    //-------------------------------------------------------------------------
    CreateDoors(&textures[3], &textures[2]);

    LCD_save(lcd);


    //-------------------------------------------------------------------------
    // allocate space for a background image and fill the lower part of the
    // light plane with 0xff to simulate a floor ...
    //-------------------------------------------------------------------------
    if (!(bg_img = (unsigned char*)malloc(1152*2))) {
        FAT_UnloadEngine();
        FAT_UnlockHandleOfFile(texfile);
        ST_helpMsg("ERROR: out of mem");
        return;
    }
    memset(bg_img,0x00,1152*2);

    dest = (unsigned short*)bg_img;
    for (i=0;i<6;i++) {
        for (j=0;j<48;j++) *dest++ = 0x0000;
        for (j=0;j<48;j++) *dest++ = 0xffff;
    }
    fc.background = (unsigned short*)bg_img;

    InstallDummyHandlerForINTs();

    //-------------------------------------------------------------------------
    // initialize grayscales, clear both planes and set the destination plane
    // addresses for the FAT-Engine
    //-------------------------------------------------------------------------
    if (!GrayOn()) {
        free(bg_img);
        RestoreHandlerForINTs();
        FAT_UnloadEngine();
        FAT_UnlockHandleOfFile(texfile);
        ST_helpMsg("ERROR: out of mem");
        return;
    }

    //-------------------------------------------------------------------------
    // display the FAT-Engine's splashscreen
    //-------------------------------------------------------------------------
    FAT_DisplayLogo(100);

    fc.dest_plane0 = GetPlane(0)+30*2; // center vertically
    fc.dest_plane1 = GetPlane(1)+30*2; // center vertically

    //-------------------------------------------------------------------------
    // setup calculator type specials
    //-------------------------------------------------------------------------
    if (TI89) {
        display_offsetx = 0;
        display_offsety = 0;
    }
    else {
        // TI92p or V200
        display_offsetx  = 40;
        display_offsety  = 14;
        fc.dest_plane0  += (14*30+4)/2;
        fc.dest_plane1  += (14*30+4)/2;
    }

    //-------------------------------------------------------------------------
    // initialize the GUI
    //-------------------------------------------------------------------------
    InitGUI(&frames_display);

    //-------------------------------------------------------------------------
    // the main rendering loop ...
    //-------------------------------------------------------------------------
    do {
        FAT_Render(&fc);   // render image

        frame_drawn(&fc,&frames_display);  // handles frame counter drawing

        if ((key = FAT_GetPadState())) {
            if (key & PADSTATE_ESC)    break;
            if (key & PADSTATE_UPDOWN) PlayerMove(&fc,key & PADSTATE_UP,0);

            if (key & PADSTATE_LEFTRIGHT) {
                // STRAFE KEY: 2ND (TI-89) / F1 (TI-92p)
                if (key & PADSTATE_KA) {
                    PlayerMove(&fc,key&PADSTATE_RIGHT,1);
                }
                else {
                    PlayerTurn(&fc,key&PADSTATE_RIGHT,ANGLE_015/2);
                }
            }
            if (key & PADSTATE_PLUS || key & PADSTATE_MINUS) {
                UpdateDoor(&fc,key & PADSTATE_PLUS);
            }
        }
    }
    while(1);

    //-------------------------------------------------------------------------
    // some clean stuff ...
    //-------------------------------------------------------------------------
    FAT_AllKeysReleased();
    GrayOff();
    free(bg_img);
    LCD_restore(lcd);
    ST_helpMsg(FAT_GetPoweredByString());
    FAT_UnlockHandleOfFile(texfile);  // unlock texture file
    FAT_UnloadEngine();
    RestoreHandlerForINTs();
    OSqclear(kbd_queue());
}

//#############################################################################
// Revision History
//#############################################################################
//
// $Log: fatd10.c,v $
// Revision 1.5  2002/08/01 12:40:28  tnussb
// pushable walls added
//
// Revision 1.4  2002/05/08 15:35:13  tnussb
// (1) display of FAT splashscreen (FAT_DisplayLogo) added
// (2) GUI of demos completely revised
//
// Revision 1.3  2002/05/07 15:22:12  tnussb
// stupid double call of FAT_LoadEngine() fixed
//
// Revision 1.2  2002/04/30 17:41:44  tnussb
// using now defines from automatically generated headerfile indices.h for indices into
// our texture file
//
// Revision 1.1  2002/04/29 10:09:10  tnussb
// initial version
//
//
