/******************************************************************************
*
* project name:   FAT-Engine
* file name:      fatd2.c
* initial date:   11/03/2001
* author:         thomas.nussbaumer@gmx.net
*
* description:    FAT-Engine - Demo 2 (Masking Demo)
*
* $Id: fatd2.c,v 1.17 2002/05/08 15:35:13 tnussb Exp $
*
******************************************************************************/

#include "../includes/fat.h"


#define  MAP_WIDTH        16
#define  MAP_HEIGHT       16

//-------------------------------------------------------------------
// defines the used map: 0 = no wall, 1 = texture1, 2 = texture2 etc.
// NOTE: the array tex64 which gets handed over to the engine as
//       part of the FATCONFIG structure has to as much textures as
//       you have specified in this map.
//       For example: if the highest number in the map is 3 then
//       tex64 has to hold 3 textures (64x64 pixels with 2 planes)
//-------------------------------------------------------------------
char map[MAP_WIDTH*MAP_HEIGHT] = {
     1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 1,
     1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2,
     2, 0, 1, 1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 2, 0, 2,
     2, 0, 2, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 2, 0, 2,
     1, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2,
     1, 0, 2, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 2, 0, 2,
     1, 0, 1, 1, 1, 1, 1, 0, 0, 2, 1, 2, 1, 2, 0, 2,
     1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2,
     2, 0, 2, 1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 0, 0, 2,
     2, 0, 2, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 2,
     2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2,
     2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2,
     1, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 2, 0, 2,
     2, 0, 1, 1, 1, 1, 1, 0, 0, 1, 1, 2, 1, 2, 0, 2,
     2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2,
     1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 1, 2, 1, 1, 1, 1 };

//-----------------------------------------------------------------------------
// include the utility routines ...
//-----------------------------------------------------------------------------
#define NO_COUNTER_SUPPORT // we use no counters in this demo
#include "utils.c"

//-----------------------------------------------------------------------------
// include file containing the texture and genericdata indices into fatdtx
// (automatically generated by TexMaker)
//-----------------------------------------------------------------------------
#include "indices.h"

short display_offset = 0;


/*===========================================================================*/
/* copy HUD gfx data from background image into HUD arrays                   */
/*===========================================================================*/
static inline void FillHUDArrays(unsigned short* background, unsigned short* p0,unsigned short* p1) {
    short i;
    unsigned short* src1 = background;
    unsigned short* src2 = &background[1000];

    for (i=0;i<96;i++,src1+=4,src2+=4) {
        *p0++ = *src1++, *p1++ = *src2++;
        *p0++ = *src1++, *p1++ = *src2++;
        *p0++ = *src1++, *p1++ = *src2++;
        *p0++ = *src1++, *p1++ = *src2++;
        *p0++ = *src1++, *p1++ = *src2++;
        *p0++ = *src1++, *p1++ = *src2++;
    }
}


/*===========================================================================*/
/* draws the background image into the grayplanes                            */
/*===========================================================================*/
static inline void DrawBackgroundImage(unsigned long* background) {
    short i;
    unsigned long* dest0;
    unsigned long* dest1;
    unsigned long* src0 = background;
    unsigned long* src1 = &background[500];

    memset(GetPlane(0),0xff,LCD_SIZE);
    memset(GetPlane(1),0xff,LCD_SIZE);

    for (i=0;i<100;i++) {
        dest0 = (unsigned long*)(GetPlane(0)+30*i+display_offset);
        dest1 = (unsigned long*)(GetPlane(1)+30*i+display_offset);
        *dest0++ = *src0++,*dest1++ = *src1++;
        *dest0++ = *src0++,*dest1++ = *src1++;
        *dest0++ = *src0++,*dest1++ = *src1++;
        *dest0++ = *src0++,*dest1++ = *src1++;
        *dest0++ = *src0++,*dest1++ = *src1++;
    }
}


//=============================================================================
// as usual ... our main function
//=============================================================================
void _main(void) {
    PADSTATES       key    = 0;
    unsigned char*  bg_img = NULL;
    short           i,j;
    unsigned short* dest;
    TEXCONFIG       textures[2];
    LCD_BUFFER      lcd;
    unsigned short  hud_plane0[6*96];
    unsigned short  hud_plane1[6*96];
    HANDLE          texfile    = 0;
    unsigned char*  src        = NULL;
    unsigned long*  background = NULL;
    unsigned short* hud_mask   = NULL;
    short           result;

    FATCONFIG fc = {
        MAP_WIDTH,  // width of map
        MAP_HEIGHT, // height of map
        map,        // pointer to mapdata
        textures,   // pointer to textures array
        0,          // number of sprites
        NULL,       // pointer to sprites array
        96,         // start pos_x
        96,         // start pos_y
        ANGLE_EAST, // orientation
        NULL,       // destination plane0
        NULL,       // destination plane1
        NULL,       // background image
        0           // frame counter
    };

    //-------------------------------------------------------------------------
    // before we try to do anything else we have to initialize the engine
    // its only valid to proceed if FAT_LOADOKAY gets returned
    //-------------------------------------------------------------------------
    if ((result = FAT_LoadEngine()) != FAT_LOADOKAY) {
        switch(result) {
            case FAT_NOTGHOSTSPACE: ST_helpMsg("ERROR: client no exepacked!");break;
            case FAT_NOTFOUND:      ST_helpMsg("ERROR: no matching engine found");break;
            case FAT_LOCKFAILED:    ST_helpMsg("ERROR: locking engine failed");break;
            case FAT_OUTOFMEM:      ST_helpMsg("ERROR: out of memory");break;
            default:                ST_helpMsg("ERROR: unknown errorcode");break;
        }
        return;
    }

    //-------------------------------------------------------------------------
    // try to load the used textures from fatdtx ...
    //-------------------------------------------------------------------------
    if ((texfile = FAT_GetHandleOfFile("fatdtx"))) {
        if ((src = FAT_LockHandleOfFile(texfile))) {
            // load house wall textures
            if (!FAT_LoadTextures(textures,src+2,TEXIDX_HOUSE1,2)) {
                src = 0; // just to indicate something went wrong
            }
            // load background image
            if (!src || !(background=FAT_GetGenericData(src+2,GENIDX_DEMO2BACKGROUND,4000))) {
                src = 0;
            }
            // load hud mask image
            if (!src || !(hud_mask=FAT_GetGenericData(src+2,GENIDX_DEMO2HUDMASK,1152))) {
                src = 0;
            }
        }
    }

    //-------------------------------------------------------------------------
    // check if loading failed ...
    //-------------------------------------------------------------------------
    if (!texfile || !src) {
        FAT_UnloadEngine();
        if (texfile) FAT_UnlockHandleOfFile(texfile);
        ST_helpMsg("ERROR: cannot textures from fatdtx");
        return;
    }

    LCD_save(lcd);


    //-------------------------------------------------------------------------
    // allocate space for a background image and fill the lower part of the
    // light plane with 0xff to simulate a floor ...
    //-------------------------------------------------------------------------
    if (!(bg_img = (unsigned char*)malloc(1152*2))) {
        FAT_UnloadEngine();
        FAT_UnlockHandleOfFile(texfile);
        ST_helpMsg("ERROR: out of mem");
        return;
    }
    memset(bg_img,0x00,1152*2);
    dest = (unsigned short*)bg_img;
    for (i=0;i<6;i++) {
        for (j=0;j<48;j++) *dest++ = 0x0000;
        for (j=0;j<48;j++) *dest++ = 0xffff;
    }
    fc.background = (unsigned short*)bg_img;

    //-------------------------------------------------------------------------
    // install dummy handlers for INT1 and INT5
    //-------------------------------------------------------------------------
    InstallDummyHandlerForINTs();

    //-------------------------------------------------------------------------
    // initialize grayscales, draw the background image and set the destination
    // plane addresses for the FAT-Engine
    //-------------------------------------------------------------------------
    if (!GrayOn()) {
        free(bg_img);
        RestoreHandlerForINTs();
        FAT_UnloadEngine();
        FAT_UnlockHandleOfFile(texfile);
        ST_helpMsg("ERROR: out of mem");
        return;
    }

    //-------------------------------------------------------------------------
    // display the FAT-Engine's splashscreen
    //-------------------------------------------------------------------------
    FAT_DisplayLogo(100);

    //-------------------------------------------------------------------------
    // setup calculator type specials
    //-------------------------------------------------------------------------
    if (TI89) display_offset = 0;
    else      display_offset   = (14*30+4);


    DrawBackgroundImage(background);
    fc.dest_plane0 = GetPlane(0)+display_offset;
    fc.dest_plane1 = GetPlane(1)+display_offset;

    //-------------------------------------------------------------------------
    // setup HUD data (note that hud_plane0 and hud_plane1 are extracted from
    // the background image previously)
    //-------------------------------------------------------------------------
    FillHUDArrays((unsigned short*)background,hud_plane0,hud_plane1);
    FAT_SetHUD(hud_plane0,hud_plane1,hud_mask);


    //-------------------------------------------------------------------------
    // the main rendering loop ...
    //-------------------------------------------------------------------------
    do {
        FAT_Render(&fc);   // render image

        if ((key = FAT_GetPadState())) {
            if (key & PADSTATE_ESC)    break;
            if (key & PADSTATE_UPDOWN) PlayerMove(&fc,key & PADSTATE_UP,0);

            if (key & PADSTATE_LEFTRIGHT) {
                // STRAFE KEY: 2ND (TI-89) / F1 (TI-92p)
                if (key & PADSTATE_KA) {
                    PlayerMove(&fc,key&PADSTATE_RIGHT,1);
                }
                else {
                    PlayerTurn(&fc,key&PADSTATE_RIGHT,ANGLE_015/2);
                }
            }
        }
    }
    while(1);

    //-------------------------------------------------------------------------
    // some clean stuff ...
    //-------------------------------------------------------------------------
    FAT_AllKeysReleased();
    GrayOff();
    free(bg_img);
    LCD_restore(lcd);
    ST_helpMsg(FAT_GetPoweredByString());
    FAT_UnlockHandleOfFile(texfile);  // unlock texture file
    FAT_UnloadEngine();
    RestoreHandlerForINTs();
    OSqclear(kbd_queue());
}

//#############################################################################
// Revision History
//#############################################################################
//
// $Log: fatd2.c,v $
// Revision 1.17  2002/05/08 15:35:13  tnussb
// (1) display of FAT splashscreen (FAT_DisplayLogo) added
// (2) GUI of demos completely revised
//
// Revision 1.16  2002/05/07 15:22:12  tnussb
// stupid double call of FAT_LoadEngine() fixed
//
// Revision 1.15  2002/04/30 17:41:44  tnussb
// using now defines from automatically generated headerfile indices.h for indices into
// our texture file
//
// Revision 1.14  2002/04/26 13:10:49  tnussb
// if FAT_LoadEngine() fails different strings will be displayed yet
//
// Revision 1.13  2002/04/25 19:28:51  tnussb
// generic commit for engine version v1.05
//
// Revision 1.12  2002/04/24 17:41:54  tnussb
// (1) using now FAT_GetPadState() to query key presses
// (2) movement evaluation block rewritten (handles now strafing, too)
//
// Revision 1.11  2002/04/18 20:27:37  tnussb
// loading now all graphical data from file fatdtx
//
// Revision 1.10  2002/04/18 12:14:47  tnussb
// using now fatdtx texture file instead of embedded textures
//
// Revision 1.9  2002/04/16 11:36:37  tnussb
// All calls to GrayMode() replaced by GrayOn/GrayOff calls
//
// Revision 1.8  2002/04/08 20:18:16  tnussb
// using now signed char (instead of unsigned char) for map
//
// Revision 1.7  2002/03/28 11:16:53  tnussb
// using now FAT_SetHUD instead of own masking
//
// Revision 1.6  2002/03/26 11:56:24  tnussb
// all warnings reported by -Wall fixed
//
// Revision 1.5  2002/03/11 09:17:41  tnussb
// USE_V200 added
//
// Revision 1.4  2002/03/07 11:14:18  tnussb
// restores now previously set font at the end
//
// Revision 1.3  2002/03/06 12:32:52  tnussb
// (1) InstallDummyHandleForINT1(), RestoreHandlerForINT1() renamed, because AUTO_INT5 will be redirected now, too
// (2) don't use SAVE_SCREEN anymore, but use own screenbackup to enable writting to the statusline
//
// Revision 1.2  2002/03/01 11:09:55  tnussb
// generic commit for sourcecode release v0.81
//
// Revision 1.1  2001/03/11 19:16:41  Thomas Nussbaumer
// initial version
//
