/******************************************************************************
*
* project name:   FAT-Engine
* file name:      fatd3.c
* initial date:   22/03/2001
* authors:        thomas.nussbaumer@gmx.net
*                 Scott Noveck (noveck@pluto.njcc.com)
*
* ("restrict player to come not too close" code in PlayerMove() developed by
*  Scott Noveck)
*
*
* description:    FAT-Engine - Demo 3
*
* demonstrates: (1) Distance Fading
*               (2) Lightning Effect
*               (3) "come not too close to a wall" effect including autosliding
*                   if characters stucks
*
* $Id: fatd3.c,v 1.20 2002/05/08 15:35:13 tnussb Exp $
*
******************************************************************************/

#include "../includes/fat.h"

#define  MAP_WIDTH        16
#define  MAP_HEIGHT       16

//-------------------------------------------------------------------
// defines the used map: 0 = no wall, 1 = texture1, 2 = texture2 etc.
// NOTE: the array tex64 which gets handed over to the engine as
//       part of the FATCONFIG structure has to as much textures as
//       you have specified in this map.
//       For example: if the highest number in the map is 3 then
//       tex64 has to hold 3 textures (64x64 pixels with 2 planes)
//-------------------------------------------------------------------

char map[MAP_WIDTH*MAP_HEIGHT] = {
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
     1, 0, 0, 2, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
     1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 0, 1, 1, 1, 0, 1,
     1, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 1,
     1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1,
     1, 0, 1, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 1,
     1, 0, 1, 1, 1, 1, 1, 0, 0, 1, 1, 1, 0, 1, 1, 1,
     1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1,
     1, 0, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 2, 1, 0, 1,
     1, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 1,
     1, 0, 1, 0, 0, 0, 2, 0, 0, 1, 0, 1, 0, 1, 0, 1,
     1, 0, 1, 0, 0, 0, 1, 0, 0, 1, 1, 1, 0, 1, 0, 1,
     1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 1,
     1, 0, 1, 1, 1, 1, 1, 0, 1, 0, 1, 1, 1, 1, 0, 1,
     1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1,
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 };


//-----------------------------------------------------------------------------
// include the utility routines ...
//-----------------------------------------------------------------------------
#define NO_PLAYER_MOVE     // we don't use the default movement code
#define NO_COUNTER_SUPPORT // we use no counters in this demo
#include "utils.c"

//-----------------------------------------------------------------------------
// include file containing the texture and genericdata indices into fatdtx
// (automatically generated by TexMaker)
//-----------------------------------------------------------------------------
#include "indices.h"


short display_offset  = 0; // will be set depending on calctype (TI89 or TI92p)
short display_offsetx = 0; // will be set depending on calctype (TI89 or TI92p)
short display_offsety = 0; // will be set depending on calctype (TI89 or TI92p)


//=============================================================================
// load background image
//=============================================================================
static inline void LoadPic(unsigned long* gui) {
    static unsigned long empty[] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    unsigned long* src1;
    unsigned long* src2;
    unsigned long* dest1;
    unsigned long* dest2;
    short i;

    src1 = gui;
    src2 = &gui[500];
    for (i=0;i<100;i++) {
        dest1 = (unsigned long*)(GetPlane(0)+(i<<5)-(i<<1)+display_offset);
        dest2 = (unsigned long*)(GetPlane(1)+(i<<5)-(i<<1)+display_offset);
        *dest1++=*src1++,*dest1++=*src1++,*dest1++=*src1++,*dest1++=*src1++,*dest1++=*src1++;
        *dest2++=*src2++,*dest2++=*src2++,*dest2++=*src2++,*dest2++=*src2++,*dest2++=*src2++;
    }

    // the tiny FAT logo goes to the upper right corner ...
    Sprite32(158-32+display_offsetx,display_offsety+2,16,empty,GetPlane(0),SPRT_AND);
    Sprite32(158-32+display_offsetx,display_offsety+2,16,FAT_GetTinyLogo(),GetPlane(0),SPRT_OR);
    Sprite32(158-32+display_offsetx,display_offsety+2,16,empty,GetPlane(1),SPRT_AND);
    Sprite32(158-32+display_offsetx,display_offsety+2,16,FAT_GetTinyLogo()+16,GetPlane(1),SPRT_OR);
}


short dont_modify = 0;

//=============================================================================
// our Predraw callback which simulates the fat-in-the-distance effect by
// modifying the textures used to render the view
//=============================================================================
void __attribute__((__stkparm__)) PredrawCallback(FATCONFIG* fc,DEPTHDATA* dd) {
    short i;

    fc = fc; // just to make the compiler happy. otherwise it will complain
             // about the unused parameter fc

    if (dont_modify) return;
    for (i=0;i<ANGLE_060;i++) {
        if (dd->distance[i] < 76) continue;
        if (dd->distance[i] < 100) {
            dd->lightdata[i<<1]     = 0xffffffffL;
            dd->lightdata[(i<<1)+1] = 0xffffffffL;
        }
        else if (dd->distance[i] < 170) {
            dd->darkdata[i<<1]     = 0xffffffffL;
            dd->darkdata[(i<<1)+1] = 0xffffffffL;
        }
        else {
            dd->lightdata[i<<1]     = 0xffffffffL;
            dd->lightdata[(i<<1)+1] = 0xffffffffL;
            dd->darkdata[i<<1]      = 0xffffffffL;
            dd->darkdata[(i<<1)+1]  = 0xffffffffL;
        }
    }
}


//=============================================================================
// our PreHUD callback which inverts the data of rendered view if don't modify
// is not set.
// NOTE that the data in array plane0 and plane1 are very specially arranged.
// Each array contains the graphics data of 6 vertical strips each 16 pixels
// wide (1 short). First comes the 96 shorts of the first strip, then the 96
// shorts of the second strip and so on.
// Well - in our case we doesn't bother about this arrangement. We just invert
// the content of the buffers.
//=============================================================================
void __attribute__((__stkparm__)) PreHUDCallback(FATCONFIG* fc,unsigned short* plane0,unsigned short* plane1) {
    short i;
    if (!dont_modify) return;
    fc = fc; // to make the compiler happy ...
    for (i=0;i<6*96;i++,plane0++,plane1++) *plane0 = ~*plane0,*plane1 = ~*plane1;
}


//=============================================================================
// returns 0 if move would bring the character too near to a wall
//=============================================================================
short ValidMove(FATCONFIG* fc, short x, short y) {
    #define DIST_FROM_WALL 15

    // if we would come too near to a solid square...
    if ((*(map+((y+DIST_FROM_WALL)>>6)*fc->map_width+((x+DIST_FROM_WALL)>>6))==1) ||
        (*(map+((y-DIST_FROM_WALL)>>6)*fc->map_width+((x-DIST_FROM_WALL)>>6))==1) ||
        (*(map+((y+DIST_FROM_WALL)>>6)*fc->map_width+((x-DIST_FROM_WALL)>>6))==1) ||
        (*(map+((y-DIST_FROM_WALL)>>6)*fc->map_width+((x+DIST_FROM_WALL)>>6))==1))
    {
       return 0;        // false
    }
    return 1;           // true
}


//=============================================================================
// move player forward or backward depending on player's angle ...
//
// special player move routine (takes doors into account) AND don't allows
// the user to come too near to a wall, but try to slide automatically.
//=============================================================================
void PlayerMove(FATCONFIG* fc,int forward) {
    short stepx   = FAT_Cos16384(fc->cam_orientation) >> 11;
    short stepy   = FAT_Sin16384(fc->cam_orientation) >> 11;
    short newxpos = fc->cam_xpos + ((forward)? stepx: -stepx);
    short newypos = fc->cam_ypos + ((forward)? stepy: -stepy);
    short newxmod = newxpos >> 6;
    short newymod = newypos >> 6;

    //-------------------------------------------------------------------------
    // check if we are still within the map
    //-------------------------------------------------------------------------
    if (newxmod > fc->map_width-1  ||
        newymod > fc->map_height-1 ||
        newxmod < 0 || newymod < 0)
    {
       return;
    }

    if (!ValidMove(fc, newxpos, newypos)) {  // if new position is invalid ...
        // ...but old position is OK, we're against a wall to east or west
        if (ValidMove(fc, fc->cam_xpos, newypos)) {
            // don't change angle if movement is perpendicular to the wall
            if (stepy != 0) {
                // if sign(stepx)==sign(stepy), rotate clockwise = turn right
                if ((stepx>0)==(stepy>0)) PlayerTurn(fc,1,ANGLE_015/2);
                else                      PlayerTurn(fc,0,ANGLE_015/2);
                // still allow movement along y = parallel to wall
                fc->cam_ypos = newypos;
            }
        }
        // ...but (new x, old y) is OK, we're against a wall to north or south
        if (ValidMove(fc, newxpos, fc->cam_ypos)) {
            // don't change angle if movement is perpendicular to the wall
            if (stepx != 0) {
               // if sign(stepx)==sign(stepy), rotate counterclockwise = turn left
               if ((stepx>0)==(stepy>0)) PlayerTurn(fc,0,ANGLE_015/2);
               else                      PlayerTurn(fc,1,ANGLE_015/2);
               // still allow movement along x = parallel to wall
               fc->cam_xpos = newxpos;
            }
        }
        // ...otherwise we're in a corner, so do nothing
        return;
    }

    //-------------------------------------------------------------------------
    // everything okay -> let's set the new position
    //-------------------------------------------------------------------------
    fc->cam_xpos = newxpos;
    fc->cam_ypos = newypos;
}


/*===========================================================================*/
/* copy HUD gfx data from background image into HUD arrays                   */
/*===========================================================================*/
static inline void FillHUDArrays(unsigned short* gui,unsigned short* p0,unsigned short* p1) {
    short i;
    unsigned short* src1 = gui;
    unsigned short* src2 = &gui[1000];

    for (i=0;i<96;i++,src1+=4,src2+=4) {
        *p0++ = *src1++, *p1++ = *src2++;
        *p0++ = *src1++, *p1++ = *src2++;
        *p0++ = *src1++, *p1++ = *src2++;
        *p0++ = *src1++, *p1++ = *src2++;
        *p0++ = *src1++, *p1++ = *src2++;
        *p0++ = *src1++, *p1++ = *src2++;
    }
}


//=============================================================================
// as usual ... our main function
//=============================================================================
void _main(void) {
    PADSTATES       key = 0;
    unsigned short* bg_img;
    short           i,j;
    unsigned short* dest;
    TEXCONFIG       textures[2];
    LCD_BUFFER      lcd;
    unsigned short  hud_plane0[6*96];       // will hold the hud plane0 data
    unsigned short  hud_plane1[6*96];       // will hold the hud plane1 data
    HANDLE          texfile = 0;
    unsigned char*  src     = NULL;
    short           oldfont;
    unsigned long*  gui        = NULL;
    unsigned short* hud_mask   = NULL;
    unsigned short* background = NULL;
    short           result;

    FATCONFIG fc = {
        MAP_WIDTH,
        MAP_HEIGHT,
        map,
        textures,
        0,
        NULL,
        96,   // start pos_x
        96,   // start pos_y
        0,    // orientation
        NULL, // destination plane0
        NULL, // destination plane1
        NULL, // background image
        0     // frame counter
    };

    //-------------------------------------------------------------------------
    // before we try to do anything else we have to initialize the engine
    // its only valid to proceed if FAT_LOADOKAY gets returned
    //-------------------------------------------------------------------------
    if ((result = FAT_LoadEngine()) != FAT_LOADOKAY) {
        switch(result) {
            case FAT_NOTGHOSTSPACE: ST_helpMsg("ERROR: client no exepacked!");break;
            case FAT_NOTFOUND:      ST_helpMsg("ERROR: no matching engine found");break;
            case FAT_LOCKFAILED:    ST_helpMsg("ERROR: locking engine failed");break;
            case FAT_OUTOFMEM:      ST_helpMsg("ERROR: out of memory");break;
            default:                ST_helpMsg("ERROR: unknown errorcode");break;
        }
        return;
    }

    //-------------------------------------------------------------------------
    // try to load the used textures from fatdtx ...
    //-------------------------------------------------------------------------
    if ((texfile = FAT_GetHandleOfFile("fatdtx"))) {
        if ((src = FAT_LockHandleOfFile(texfile))) {
            // load wall and door textures
            if (!FAT_LoadTextures(textures,src+2,TEXIDX_DEMO3WALL,2)) {
                src = 0; // just to indicate something went wrong
            }
            // load gui image
            if (!src || !(gui=FAT_GetGenericData(src+2,GENIDX_DEMO3GUI,4000))) {
                src = 0;
            }

            // load hud mask image
            if (!src || !(hud_mask=FAT_GetGenericData(src+2,GENIDX_DEMO3HUDMASK,1152))) {
                src = 0;
            }
            // load background image
            if (!src || !(background=FAT_GetGenericData(src+2,GENIDX_DEMO3BACKGROUND,2304))) {
                src = 0;
            }
        }
    }

    //-------------------------------------------------------------------------
    // check if loading failed ...
    //-------------------------------------------------------------------------
    if (!texfile || !src) {
        FAT_UnloadEngine();
        if (texfile) FAT_UnlockHandleOfFile(texfile);
        ST_helpMsg("ERROR: cannot textures from fatdtx");
        return;
    }

    LCD_save(lcd);


    //-------------------------------------------------------------------------
    // allocate space for a background image and copy the used image into it
    //-------------------------------------------------------------------------
    if (!(bg_img = (unsigned short*)malloc(1152*2))) {
        FAT_UnloadEngine();
        FAT_UnlockHandleOfFile(texfile);
        ST_helpMsg("ERROR: out of mem");
        return;
    }
    memset(bg_img,0,1152*2);

    dest = bg_img;

    for (i=0;i<6;i++) {
        unsigned short* src = background+i;
        for (j=0;j<96;j++,dest++,src+=6) {
            *dest = *src;
            *(dest+576) = *(src+576);
        }
    }
    fc.background = bg_img;

    InstallDummyHandlerForINTs();

    //-------------------------------------------------------------------------
    // initialize grayscales, clear both planes and set the destination plane
    // addresses for the FAT-Engine
    //-------------------------------------------------------------------------
    if (!GrayOn()) {
        free(bg_img);
        RestoreHandlerForINTs();
        FAT_UnloadEngine();
        FAT_UnlockHandleOfFile(texfile);
        ST_helpMsg("ERROR: out of mem");
        return;
    }

    // display the FAT-Engine's splashscreen
    FAT_DisplayLogo(100);

    memset(GetPlane(0),0,LCD_SIZE);
    memset(GetPlane(1),0,LCD_SIZE);

    if (TI89) {
        display_offset  = 0;
        display_offsetx = 0;
        display_offsety = 0;
    }
    else {
        // TI92p or V200
        display_offset   = 14*30+4;
        display_offsetx  = 40;
        display_offsety  = 14;
    }

    fc.dest_plane0 = (unsigned short*)(GetPlane(0)+display_offset);
    fc.dest_plane1 = (unsigned short*)(GetPlane(1)+display_offset);

    oldfont=FontSetSys(F_4x6);

    //-------------------------------------------------------------------------
    // setup HUD data (note that hud_plane0 and hud_plane1 are extracted from
    // the background image previously)
    //-------------------------------------------------------------------------
    FillHUDArrays((unsigned short*)gui,hud_plane0,hud_plane1);
    FAT_SetHUD(hud_plane0,hud_plane1,hud_mask);

    //-------------------------------------------------------------------------
    // set predraw callback function which is called before the vertical strips
    // get rendered, but after the DEPTHDATA structure is filled ...
    //-------------------------------------------------------------------------
    FAT_SetPredrawCallback(PredrawCallback);

    //-------------------------------------------------------------------------
    // set prehud callback function which is called before the rendered view
    // is copied into the destination plane. Note that the hud is NOT applied
    // yet, when this callback is called
    //-------------------------------------------------------------------------
    FAT_SetPreHUDCallback(PreHUDCallback);

    LoadPic(gui);

    //-------------------------------------------------------------------------
    // the main rendering loop ...
    //-------------------------------------------------------------------------
    do {
        if (!(fc.frame_counter % (random(50)+1))) {
            dont_modify = 1;
            FAT_Render(&fc);
        }
        else {
            dont_modify = 0;
            FAT_Render(&fc);
        }

        if ((key = FAT_GetPadState())) {
           if (key & PADSTATE_ESC)       break;
           if (key & PADSTATE_UPDOWN)    PlayerMove(&fc,key & PADSTATE_UP);
           if (key & PADSTATE_LEFTRIGHT) PlayerTurn(&fc,key & PADSTATE_RIGHT,ANGLE_015/2);
        }
    }
    while(1);

    //-------------------------------------------------------------------------
    // some clean stuff ...
    //-------------------------------------------------------------------------
    FAT_AllKeysReleased();
    GrayOff();
    free(bg_img);
    LCD_restore(lcd);
    ST_helpMsg(FAT_GetPoweredByString());
    FAT_UnlockHandleOfFile(texfile);  // unlock texture file
    FAT_UnloadEngine();
    RestoreHandlerForINTs();
    OSqclear(kbd_queue());
    FontSetSys(oldfont);
}

//#############################################################################
// Revision History
//#############################################################################
//
// $Log: fatd3.c,v $
// Revision 1.20  2002/05/08 15:35:13  tnussb
// (1) display of FAT splashscreen (FAT_DisplayLogo) added
// (2) GUI of demos completely revised
//
// Revision 1.19  2002/05/07 15:22:13  tnussb
// stupid double call of FAT_LoadEngine() fixed
//
// Revision 1.18  2002/04/30 17:41:44  tnussb
// using now defines from automatically generated headerfile indices.h for indices into
// our texture file
//
// Revision 1.17  2002/04/26 13:10:49  tnussb
// if FAT_LoadEngine() fails different strings will be displayed yet
//
// Revision 1.16  2002/04/25 19:30:05  tnussb
// generic commit for engine version v1.05
//
// Revision 1.15  2002/04/24 17:42:51  tnussb
// (1) using now FAT_GetPadState() to query key presses
//
// Revision 1.14  2002/04/18 20:27:37  tnussb
// loading now all graphical data from file fatdtx
//
// Revision 1.13  2002/04/18 12:14:47  tnussb
// using now fatdtx texture file instead of embedded textures
//
// Revision 1.12  2002/04/17 10:03:04  tnussb
// using now macros FAT_Cos16384/FAT_Sin16384
//
// Revision 1.11  2002/04/16 11:36:37  tnussb
// All calls to GrayMode() replaced by GrayOn/GrayOff calls
//
// Revision 1.10  2002/04/08 20:18:16  tnussb
// using now signed char (instead of unsigned char) for map
//
// Revision 1.9  2002/03/28 12:20:54  tnussb
// uses now FAT_SetHUD() and FAT_SetPreHUDCallback()
//
// Revision 1.8  2002/03/26 11:56:24  tnussb
// all warnings reported by -Wall fixed
//
// Revision 1.7  2002/03/12 14:37:33  tnussb
// ValidMove() slightly improved (not checking the "inside" wall condition anymore) [Scott Noveck]
//
// Revision 1.6  2002/03/11 09:17:42  tnussb
// USE_V200 added
//
// Revision 1.5  2002/03/11 09:12:07  tnussb
// Scott's "come not too close to wall" extension added to PlayerMove()
//
// Revision 1.3  2002/03/07 11:15:37  tnussb
// completely restructured to show now also a lightning effect, a background
// image as GUI and a decoration around the FAT view.
//
// Revision 1.2  2002/03/06 12:33:23  tnussb
// (1) OutputTitle completely rewritten (uses now new exported "functions" of FAT-Engine)
// (2) InstallDummyHandleForINT1(), RestoreHandlerForINT1() renamed, because AUTO_INT5 will be redirected now, too
// (3) don't use SAVE_SCREEN anymore, but use own screenbackup to enable writting to the statusline
//
// Revision 1.1  2002/03/01 11:09:55  tnussb
// generic commit for sourcecode release v0.81
//
//
