/******************************************************************************
*
* project name:   FAT-Engine
* file name:      fatd4.c
* initial date:   22/03/2001
* author:         thomas.nussbaumer@gmx.net
*
* description:    FAT-Engine - Demo 4 (Simple Sprite Demo)
*
* $Id: fatd4.c,v 1.19 2002/05/08 15:35:13 tnussb Exp $
*
******************************************************************************/

#include "../includes/fat.h"

#define  MAP_WIDTH        16
#define  MAP_HEIGHT       16


//-------------------------------------------------------------------
// defines the used map: 0 = no wall, 1 = texture1, 2 = texture2 etc.
// NOTE: the array tex64 which gets handed over to the engine as
//       part of the FATCONFIG structure has to as much textures as
//       you have specified in this map.
//       For example: if the highest number in the map is 3 then
//       tex64 has to hold 3 textures (64x64 pixels with 2 planes)
//-------------------------------------------------------------------

char map[MAP_WIDTH*MAP_HEIGHT] = {
     2, 1, 2, 2, 1, 2, 2, 2, 2, 2, 2, 1, 2, 2, 1, 2,
     2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
     2, 0, 2, 2, 2, 2, 2, 0, 0, 2, 2, 2, 2, 2, 0, 2,
     2, 0, 2, 0, 0, 0, 2, 0, 0, 2, 0, 0, 0, 1, 0, 2,
     2, 0, 1, 0, 3, 0, 0, 0, 0, 0, 0, 3, 0, 2, 0, 2,
     2, 0, 2, 0, 0, 0, 2, 0, 0, 2, 0, 0, 0, 1, 0, 2,
     2, 0, 0, 2, 2, 2, 0, 0, 0, 0, 2, 2, 2, 0, 0, 2,
     2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2,
     2, 0, 2, 2, 2, 2, 2, 0, 0, 2, 2, 1, 2, 2, 0, 2,
     2, 0, 1, 0, 0, 0, 2, 0, 0, 2, 0, 0, 0, 1, 0, 2,
     2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2,
     2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2,
     2, 0, 1, 0, 0, 0, 2, 0, 0, 2, 0, 0, 0, 1, 0, 2,
     2, 0, 2, 2, 2, 2, 2, 0, 0, 2, 2, 2, 2, 2, 0, 2,
     2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
     2, 1, 2, 2, 1, 2, 2, 2, 2, 2, 2, 1, 2, 2, 1, 2 };

//-----------------------------------------------------------------------------
// include the utility routines ...
//-----------------------------------------------------------------------------
#include "utils.c"

//-----------------------------------------------------------------------------
// include file containing the texture and genericdata indices into fatdtx
// (automatically generated by TexMaker)
//-----------------------------------------------------------------------------
#include "indices.h"


short display_offsetx = 0; // will be set depending on calctype (TI89 or TI92p)
short display_offsety = 0; // will be set depending on calctype (TI89 or TI92p)


//=============================================================================
// initializes GUI (darkgray background, tiny logo and counter)
//=============================================================================
void InitGUI(COUNTER *counter) {
    static unsigned long empty[] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

    memset(GetPlane(LIGHT_PLANE),0,LCD_SIZE);
    memset(GetPlane(DARK_PLANE),0xff,LCD_SIZE);

    Sprite32(158-32+display_offsetx,display_offsety+2,16,empty,GetPlane(0),SPRT_AND);
    Sprite32(158-32+display_offsetx,display_offsety+2,16,FAT_GetTinyLogo(),GetPlane(0),SPRT_OR);
    Sprite32(158-32+display_offsetx,display_offsety+2,16,empty,GetPlane(1),SPRT_AND);
    Sprite32(158-32+display_offsetx,display_offsety+2,16,FAT_GetTinyLogo()+16,GetPlane(1),SPRT_OR);

    InitCounter(counter,158-6*5+display_offsetx,99-10+display_offsety,5,0);
    DrawCounter(counter);
}


//=============================================================================
// output frame count ever 10th frame
//=============================================================================
void frame_drawn(FATCONFIG* fc,COUNTER* counter) {
    short i;
    if (fc->frame_counter%10==0) {
        // we starting incrementing at 2nd position and not on first!
        for (i=counter->length-2;i>=0;i--) {

            counter->newcount[i] = counter->newcount[i]+1;
            if (counter->newcount[i] < 10) break;
            else                           counter->newcount[i]=0;
        }
        DrawCounter(counter);
    }
}


//=============================================================================
// as usual ... our main function
//=============================================================================
void _main(void) {
    PADSTATES       key    = 0;
    unsigned char*  bg_img = NULL;
    short           i,j;
    unsigned short* dest;
    TEXCONFIG       textures[3];
    FATSPRITE       sprite;
    short           direction = 1;
    LCD_BUFFER      lcd;
    HANDLE          texfile = 0;
    unsigned char*  src = 0;
    short           result;
    COUNTER         frames_display;

    FATCONFIG fc = {
        MAP_WIDTH,  // width of map
        MAP_HEIGHT, // height of map
        map,        // pointer to map data
        textures,   // pointer to textures array
        1,          // number of sprites
        &sprite,    // array of sprites (in this case just one)
        96,         // start pos_x
        96,         // start pos_y
        ANGLE_EAST, // orientation
        NULL,       // destination plane0
        NULL,       // destination plane1
        NULL,       // background image
        0           // frame counter
    };

    //-------------------------------------------------------------------------
    // before we try to do anything else we have to initialize the engine
    // its only valid to proceed if FAT_LOADOKAY gets returned
    //-------------------------------------------------------------------------
    if ((result = FAT_LoadEngine()) != FAT_LOADOKAY) {
        switch(result) {
            case FAT_NOTGHOSTSPACE: ST_helpMsg("ERROR: client no exepacked!");break;
            case FAT_NOTFOUND:      ST_helpMsg("ERROR: no matching engine found");break;
            case FAT_LOCKFAILED:    ST_helpMsg("ERROR: locking engine failed");break;
            case FAT_OUTOFMEM:      ST_helpMsg("ERROR: out of memory");break;
            default:                ST_helpMsg("ERROR: unknown errorcode");break;
        }
        return;
    }

    //-------------------------------------------------------------------------
    // try to load the used textures from fatdtx ...
    //-------------------------------------------------------------------------
    if ((texfile = FAT_GetHandleOfFile("fatdtx"))) {
        if ((src = FAT_LockHandleOfFile(texfile))) {
            // load house wall textures
            if (!FAT_LoadTextures(textures,src+2,TEXIDX_HOUSE1,2)) {
                src = 0; // just to indicate something went wrong
            }
            // load texture of TICT logo into index 3 (used as sprite and as wall texture!)
            if (src) {
                if (!FAT_LoadTextures(&textures[2],src+2,TEXIDX_TICTLOGO,1)) {
                    src = 0; // just to indicate something went wrong
                }
            }
        }
    }

    //-------------------------------------------------------------------------
    // check if loading failed ...
    //-------------------------------------------------------------------------
    if (!texfile || !src) {
        FAT_UnloadEngine();
        if (texfile) FAT_UnlockHandleOfFile(texfile);
        ST_helpMsg("ERROR: cannot textures from fatdtx");
        return;
    }

    LCD_save(lcd);

    //-------------------------------------------------------------------------
    // initialize sprite structure and setup its fields
    //-------------------------------------------------------------------------
    FAT_InitSprite(&sprite);
    sprite.drawmode    = SPRITEMODE_SIZE64;
    sprite.nr_textures = 1;
    sprite.textures    = &textures[2];
    sprite.orientation = 0;
    sprite.xpos        = 96+32+96*2;
    sprite.ypos        = 96;

    //-------------------------------------------------------------------------
    // allocate space for a background image and fill the lower part of the
    // light plane with 0xff to simulate a floor ...
    //-------------------------------------------------------------------------
    if (!(bg_img = (unsigned char*)malloc(1152*2))) {
        FAT_UnloadEngine();
        FAT_UnlockHandleOfFile(texfile);
        ST_helpMsg("ERROR: out of mem");
        return;
    }
    memset(bg_img,0x00,1152*2);

    dest = (unsigned short*)bg_img;
    for (i=0;i<6;i++) {
        for (j=0;j<48;j++) *dest++ = 0x0000;
        for (j=0;j<48;j++) *dest++ = 0xffff;
    }

    fc.background = (unsigned short*)bg_img;

    InstallDummyHandlerForINTs();

    //-------------------------------------------------------------------------
    // initialize grayscales, clear both planes and set the destination plane
    // addresses for the FAT-Engine
    //-------------------------------------------------------------------------
    if (!GrayOn()) {
        FAT_UnloadEngine();
        FAT_UnlockHandleOfFile(texfile);
        free(bg_img);
        RestoreHandlerForINTs();
        ST_helpMsg("ERROR: out of mem");
        return;
    }

    //-------------------------------------------------------------------------
    // display the FAT-Engine's splashscreen
    //-------------------------------------------------------------------------
    FAT_DisplayLogo(100);

    fc.dest_plane0 = GetPlane(0)+30*2;  // center vertically
    fc.dest_plane1 = GetPlane(1)+30*2;  // center vertically

    //-------------------------------------------------------------------------
    // setup calculator type specials
    //-------------------------------------------------------------------------
    if (TI89) {
        display_offsetx = 0;
        display_offsety = 0;
    }
    else {
        // TI92p or V200
        display_offsetx  = 40;
        display_offsety  = 14;
        fc.dest_plane0  += (14*30+4)/2;
        fc.dest_plane1  += (14*30+4)/2;
    }

    //-------------------------------------------------------------------------
    // initialize the GUI
    //-------------------------------------------------------------------------
    InitGUI(&frames_display);

    //-------------------------------------------------------------------------
    // the main rendering loop ...
    //-------------------------------------------------------------------------
    do {
        FAT_Render(&fc);   // render image
        // check position of sprite. if the sprite reaches a border, modify its
        // moving direction
        if (fc.sprites[0].xpos > (MAP_WIDTH-2)*64) direction = -1;
        else if (fc.sprites[0].xpos < 96)          direction = 1;

        // move sprite
        fc.sprites[0].xpos += direction<<3;
        frame_drawn(&fc,&frames_display);  // handles frame counter drawing

        if ((key = FAT_GetPadState())) {
           if (key & PADSTATE_ESC)    break;
           if (key & PADSTATE_UPDOWN) PlayerMove(&fc,key & PADSTATE_UP,0);

           if (key & PADSTATE_LEFTRIGHT) {
               // STRAFE KEY: 2ND (TI-89) / F1 (TI-92p)
               if (key & PADSTATE_KA) {
                   PlayerMove(&fc,key&PADSTATE_RIGHT,1);
               }
               else {
                   PlayerTurn(&fc,key&PADSTATE_RIGHT,ANGLE_015/2);
               }
           }
        }
    }
    while(1);

    //-------------------------------------------------------------------------
    // some clean stuff ...
    //-------------------------------------------------------------------------
    FAT_AllKeysReleased();
    GrayOff();
    free(bg_img);
    LCD_restore(lcd);
    ST_helpMsg(FAT_GetPoweredByString());
    FAT_UnlockHandleOfFile(texfile);  // unlock texture file
    FAT_UnloadEngine();
    RestoreHandlerForINTs();
    OSqclear(kbd_queue());
}

//#############################################################################
// Revision History
//#############################################################################
//
// $Log: fatd4.c,v $
// Revision 1.19  2002/05/08 15:35:13  tnussb
// (1) display of FAT splashscreen (FAT_DisplayLogo) added
// (2) GUI of demos completely revised
//
// Revision 1.18  2002/05/07 15:22:13  tnussb
// stupid double call of FAT_LoadEngine() fixed
//
// Revision 1.17  2002/04/30 17:41:44  tnussb
// using now defines from automatically generated headerfile indices.h for indices into
// our texture file
//
// Revision 1.16  2002/04/26 13:10:49  tnussb
// if FAT_LoadEngine() fails different strings will be displayed yet
//
// Revision 1.15  2002/04/25 19:41:44  tnussb
// generic commit for engine version v1.05
//
// Revision 1.14  2002/04/24 17:41:54  tnussb
// (1) using now FAT_GetPadState() to query key presses
// (2) movement evaluation block rewritten (handles now strafing, too)
//
// Revision 1.13  2002/04/18 12:14:47  tnussb
// using now fatdtx texture file instead of embedded textures
//
// Revision 1.12  2002/04/17 10:03:49  tnussb
// (1) using now macros FAT_Cos16384/FAT_Sin16384
// (2) sprite rendering debug mode added (define DEBUG_VERSION)
//
// Revision 1.11  2002/04/16 11:36:37  tnussb
// All calls to GrayMode() replaced by GrayOn/GrayOff calls
//
// Revision 1.10  2002/04/15 10:00:12  tnussb
// using now STRIP_NORMAL instead of 0 for mirrored field structure TEXSTRIP
//
// Revision 1.9  2002/04/08 20:18:16  tnussb
// using now signed char (instead of unsigned char) for map
//
// Revision 1.8  2002/03/26 11:56:24  tnussb
// all warnings reported by -Wall fixed
//
// Revision 1.7  2002/03/12 14:51:07  tnussb
// sprite texture is used as wall, too (to see that transparency is only
// used for sprites)
//
// Revision 1.6  2002/03/12 14:36:19  tnussb
// uses now FAT_InitSprite() and setup drawmode of sprite correctly
//
// Revision 1.5  2002/03/11 09:17:42  tnussb
// USE_V200 added
//
// Revision 1.4  2002/03/08 13:36:19  tnussb
// necessary changes for new sprite support (sprite.nr_textures)
//
// Revision 1.3  2002/03/07 11:16:10  tnussb
// restores now previously set font at the end
//
// Revision 1.2  2002/03/06 12:34:26  tnussb
// (1) OutputTitle completely rewritten (uses now new exported "functions" of FAT-Engine)
// (2) InstallDummyHandleForINT1(), RestoreHandlerForINT1() renamed, because AUTO_INT5 will be redirected now, too
// (3) don't use SAVE_SCREEN anymore, but use own screenbackup to enable writting to the statusline
//
// Revision 1.1  2002/03/01 11:09:55  tnussb
// generic commit for sourcecode release v0.81
//
//
//
