/******************************************************************************
*
* project name:    FAT-SDK
* file name:       fatload.c
* initial date:    03/05/2002
* author:          thomas.nussbaumer@gmx.net
* description:     start utility for exepacked FAT-Demos
*
* (based on the code of ttstart from the TIGCC Tools Suite)
*
* $Id: fatload.c,v 1.2 2002/05/22 09:41:55 tnussb Exp $
*
******************************************************************************/

#define USE_TI89       // compile for TI89
#define USE_TI92P      // compile for TI92p
#define USE_V200       // compile for V200

#define C89_92(x,y) ((TI89)?(x):(y))
//-----------------------------------------------------------------------------
// Undocumented macro in TIGCC v0.95 and above: Do not create global imports
// for startup code. This macro is defined exclusively for ttstart; do not use
// it in your own programs. In ttstart, it only suppresses a lot of warnings.
//-----------------------------------------------------------------------------
#define _NO_INCLUDE_PATCH

#define NO_EXIT_SUPPORT             // the launcher doesn't need exit support
#define NO_CALC_DETECT              // it doesn't need calc detection, either
#define OPTIMIZE_ROM_CALLS          // generates smaller code by reserving a5
#define ENABLE_ERROR_RETURN         // to return errors to AMS.
#ifndef TIGCC_PSTARTER
#endif

#include <version.h>
#include <statline.h>
#include <alloc.h>
#include <vat.h>
#include <string.h>
#include <system.h>
#include <error.h>
#include <estack.h>
#include <compat.h>
#include <flash.h>
#include <graph.h>
#include <menus.h>
#include <kbd.h>

#define ONLY_PPG
#undef TIGCC_PSTARTER
#define REENTRANT
#undef LEAKWATCH

// Starting from TIGCC v0.95, we use our own startup code.
#if __TIGCC_VERSION__ >= 95

// ttst121: from void* (would generate warnings with -Wpointer-arith).
unsigned char *real_saved_sp;
#define saved_sp (real_saved_sp-4)
#define SAVED_SP_OFFSET 2 // longwords

asm(".section _st1\n"
"entry_point:"
// OPTIMIZE_ROM_CALLS support.
"	pea.l (%a5)\n"
"	move.l 0xC8,%a5\n"
// Save the stack pointer.
"	lea.l real_saved_sp(%pc),%a0\n"
"	move.l %a7,(%a0)\n"
// Call main_function.
"	bsr.s main_function\n"
// OPTIMIZE_ROM_CALLS support.
"	move.l (%sp)+,%a5\n"
// Return.
"	rts\n"
".text");

extern unsigned char entry_point;

#else

//-----------------------------------------------------------------------------
// DIRTY: This is needed to get everything to work with "complex main"s
//        (Kevin Kofler)
//-----------------------------------------------------------------------------
#if __TIGCCLIB_VERSION__ != 250
#warning Wrong TIGCCLIB version! Do not expect ttstart to work!
#endif
asm("/* Include Patch: save_the_sp */");
extern void *saved_sp asm("__save__sp__");
#ifdef TIGCC_PSTARTER
#define SAVED_SP_OFFSET 18 // longwords
#else
#define SAVED_SP_OFFSET 22 // longwords
#endif

#endif

#define fatal(s) {ST_helpMsg(s); return;}


typedef struct {
    unsigned char  osize_lo;   // original size lowbyte
    unsigned char  osize_hi;   // original size highbyte
    unsigned char  magic1;     // must be equal to TTUNPACK_MAGIC1
    unsigned char  magic2;     // must be equal to TTUNPACK_MAGIC2
    unsigned char  csize_lo;   // compressed size lowbyte
    unsigned char  csize_hi;   // compressed size lowbyte
    unsigned char  esc1;       // escape >> (8-escBits)
    unsigned char  notused3;
    unsigned char  notused4;
    unsigned char  esc2;       // escBits
    unsigned char  gamma1;     // maxGamma + 1
    unsigned char  gamma2;     // (1<<maxGamma)
    unsigned char  extralz;    // extraLZPosBits
    unsigned char  notused1;
    unsigned char  notused2;
    unsigned char  rleentries; // rleUsed
} TTUNPACK_HEADER;

#define TTUNPACK_MAGIC1 0x54
#define TTUNPACK_MAGIC2 0x50

#define ttunpack_size(_p_)  ((unsigned short)(((TTUNPACK_HEADER*)(_p_))->osize_lo | (((TTUNPACK_HEADER*)(_p_))->osize_hi << 8)))
#define ttunpack_valid(_p_) (((TTUNPACK_HEADER*)(_p_))->magic1 == TTUNPACK_MAGIC1 && ((TTUNPACK_HEADER*)(_p_))->magic2 == TTUNPACK_MAGIC2)

#define TTUNPACK_OKAY             0
#define TTUNPACK_NOESCFOUND     248
#define TTUNPACK_ESCBITS        249
#define TTUNPACK_MAXGAMMA       250
#define TTUNPACK_EXTRALZP       251
#define TTUNPACK_NOMAGIC        252
#define TTUNPACK_OUTBUFOVERRUN  253
#define TTUNPACK_LZPOSUNDERRUN  254

#define ttunpack_decompress ((__attribute__((__stkparm__)) unsigned short(*)(unsigned char*,unsigned char*))(unsigned short[]){\
0x206f,0x0004,0x226f,0x0008,0x48e7,0x010c,0x303c,0x0200,0x4e41,0x3f00,\
0x610e,0x3200,0x301f,0x4e41,0x3001,0x4cdf,0x3080,0x4e75,0x4e56,\
0xfff4,0x48e7,0x1e30,0x2a09,0x0c28,0x0054,0x0002,0x6708,0x303c,0x00fc,\
0x6000,0x021e,0x0c28,0x0050,0x0003,0x66f0,0x0c28,0x0008,0x000a,0x6708,\
0x303c,0x00fa,0x6000,0x0206,0x0c28,0xff80,0x000b,0x66f0,0x4240,0x1028,\
0x0009,0x3d40,0xfff8,0x303c,0x00f9,0x0c6e,0x0008,0xfff8,0x6200,0x01e6,\
0x4240,0x1028,0x000c,0x3d40,0xfff6,0x303c,0x00fb,0x0c6e,0x0004,0xfff6,\
0x6200,0x01ce,0x4240,0x1028,0x0006,0x3640,0x7008,0x906e,0xfff8,0x3d40,\
0xfffa,0x43e8,0x000f,0x2d49,0xfffc,0x4fef,0xfefc,0x2a4f,0x327c,0x0001,\
0x4217,0x4202,0x1002,0x4880,0x7201,0xe169,0x6f0c,0x3001,0x1b82,0x9000,\
0x5249,0x5340,0x66f6,0x5202,0x0c02,0x0007,0x6fe2,0x7000,0x1028,0x000f,\
0x49f0,0x0810,0x3e3c,0x0080,0x4a6e,0xfff8,0x6600,0x0162,0x380b,0xb84b,\
0x6718,0x302e,0xfffa,0xe16c,0x2445,0x5285,0x3200,0x41fa,0x01c6,0x4e90,\
0x8800,0x1484,0x60da,0x426e,0xfff4,0x45fa,0x016a,0x4e92,0x3c00,0x0c40,\
0x0001,0x6600,0x00e2,0x4240,0x1014,0xc047,0x6700,0x00a4,0xe21f,0x6402,\
0x528c,0x4240,0x1014,0xc047,0x676a,0xe21f,0x6402,0x528c,0x4e92,0x3800,\
0x0c40,0x007f,0x6240,0x4e92,0x0c40,0x001f,0x6224,0x0280,0x0000,0xffff,\
0x206e,0xfffc,0x1030,0x0800,0x0240,0x00ff,0x4242,0x2245,0x5285,0x12c0,\
0x5242,0xb842,0x64f4,0x6000,0xff74,0x3600,0x0643,0xffe0,0xe74b,0x7203,\
0x41fa,0x0148,0x4e90,0x8043,0x60da,0x0644,0xff80,0xe34c,0x7201,0x41fa,\
0x0136,0x4e90,0x8840,0x4e92,0x5340,0xe148,0x8840,0x60a6,0xe21f,0x6402,\
0x528c,0x47fa,0x011e,0x322e,0xfff8,0x4e93,0x3600,0x302e,0xfffa,0xe16c,\
0x2445,0x5285,0x3200,0x4e93,0x8800,0x1484,0x3643,0x6000,0xff1c,0xe21f,\
0x6402,0x528c,0x41fa,0x0094,0x4e90,0x3200,0x0a41,0x00ff,0x4242,0x7000,\
0x3001,0x2045,0x91c0,0x102e,0xfff5,0xd028,0xffff,0x2245,0x5285,0x12c0,\
0x5242,0xbc42,0x64e4,0x6000,0xfee8,0x4e92,0x3600,0x5343,0x0c43,0x00fe,\
0x672a,0x4a6e,0xfff6,0x6612,0x41fa,0x0054,0x4e90,0x3200,0x0a41,0x00ff,\
0xe14b,0x8243,0x60ba,0x302e,0xfff6,0xe16b,0x3200,0x41fa,0x009a,0x4e90,\
0x8640,0x60dc,0x0c46,0x0003,0x630e,0x45fa,0x002a,0x4e92,0x3d40,0xfff4,\
0x4e92,0x608e,0x4240,0x6010,0x322e,0xfff8,0x41fa,0x0074,0x4e90,0x3800,\
0x6000,0xfe94,0x4cee,0x0c78,0xffdc,0x4e5e,0x4e75,0x101c,0xe148,0x1014,\
0x1235,0x7000,0x5201,0xe268,0x0240,0x00ff,0x4e75,0x3407,0x1014,0xc400,\
0x673c,0xd402,0x5302,0x1235,0x7000,0xc002,0xb102,0x6608,0x142c,0x0001,\
0x4602,0x5001,0x9235,0x2000,0x1401,0x5202,0x0c01,0x0007,0x6504,0x7207,\
0x7407,0x3007,0xe43f,0xbe00,0x6502,0x528c,0x6100,0x0010,0x5201,0x03c0,\
0x4e75,0x7001,0xe21f,0x6526,0x4e75,0x7000,0x5301,0x6520,0x6732,0x3007,\
0xe23f,0xbe00,0x6518,0xd040,0x5340,0xc01c,0xe148,0x1014,0x1435,0x7800,\
0xe468,0xe21f,0x6402,0x528c,0x4e75,0xd040,0x5340,0xc014,0x1435,0x7800,\
0xe428,0xe21f,0x65ec,0x4e75,0x1014,0xc007,0x56c0,0x0200,0x0001,0xe21f,\
0x65dc,0x4e75})


//-----------------------------------------------------------------------------
// The following macro calls function fptr and treats a possible RETURN_VALUE
// correctly. If a return value is available mode is set to 0, otherwise 0xFF.
//
// NOTE: the NOPs are very important due to a dirty hack in the TIGCCLIB with
//       #define RETURN_VALUE (they must be exactly after the call!!)
//
// One nop seems to be enough, that's why I (Lionel Debroux) removed one nop.
//
// Modified by Kevin Kofler to use LCD_MEM, which overlaps with the
// 0x5000-0x5FFF range, which is always execution-unprotected. Problem: this
// is not compatible with HSR... Additionnally, this may be unsafe on
// future AMS or HW versions, so we must get rid of this...
//
// Modified again by Lionel Debroux with the help of Kevin Kofler to copy
// the deprotection stub to the end of the allocated block where the
// decompressed program goes (Kevin says kernels might whine if copied to
// the beginning of the block).
//-----------------------------------------------------------------------------
#define ASM_CALL_WITH_RETURN_VALUE_HANDLING(fptr,len,mode) ({\
        ercatcherpc=__errFrame->PC;\
        /* Stupid GCC insists on computing the address in a lame way. */ \
        asm ("lea (ercatcher-middle)(%1,%2.l),%0":"=a"(__errFrame->PC)\
             :"a"(fptr),"da"((long)len):"cc");\
        memcpy((unsigned char*)(fptr)+(len),middle,92);\
        asm volatile ("movem.l %%d1-%%d7/%%a0-%%a6,-(%%sp);"\
            "pea target(%%PC);"\
            "moveq #0,%%d0;"\
            "move.w %2,%%d0;"\
            "movea.l %1,%%a3;"\
            "movea.l %%a3,%%a4;"\
            "pea 0(%1,%%d0.l);"\
            "jbra unprotect_a3_jumpto_a4;"\
            "target:"\
            "lea (%%sp,36),%%sp\n"\
            "move.l (%%sp)+,%%d0;"\
            "movem.l (%%sp)+,%%d1-%%d7/%%a0-%%a6;"\
            "move.b %%d0,%0"\
            : "=g" (mode) : "a" ((void *)(fptr)), "g" (len) : "d0","cc","memory");})

extern char middle[],ercatcher[];
extern unsigned long ercatcherpc;
asm(
/* Begin code executed at end of program */
"middle: bra.s 0f;"
"nop;"
"clr.b %d0;"
"bra.s 1f;"
"0: st.b %d0;"
"1: movea.l (%sp)+,%a4\n"
"lea entry_point-target(%a4),%a3\n"
"move.l %d0,16(%sp)\n"
"unprotect_a3_jumpto_a4:\n"
"	lea (%sp,-20),%sp\n"
"	move.l #0x820000,%d0\n"
"	move.l %d0,(%sp,12)\n"
"	move.l %d0,(%sp,16)\n"
"	move.l 20(%sp),(%sp)\n"
"	move.l 24(%sp),4(%sp)\n"
"	trap #0xC\n"
"	move.w #0x2700,%sr\n"
"	moveq.l #0xF,%d3\n"
"	pea (%a4)\n"
"	clr.w  -(%sp)\n"
"	move.l 0xAC,%a0\n"
"	jmp (%a0)\n"
"ercatcherentrypoint: .long entry_point\n"
"ercatcherpc: .long 0\n"
".even\n"
"ercatcher:\n"
"movem.l %d0-%d7/%a0-%a6,-(%sp)\n"
"move.l ercatcherentrypoint(%PC),%a3\n"
"lea.l ercatchercleanup-entry_point(%a3),%a4\n"
"jbra unprotect_a3_jumpto_a4\n"
/* End code executed at end of program */

"ercatchercleanup:\n"
"lea (%sp,20),%sp\n"
"movem.l (%sp)+,%d0-%d7/%a0-%a6\n"
"move.l ercatcherpc(%PC),-(%a7)\n"
"rts\n"
);

#ifdef USE_OWN_EXPATCH

//-----------------------------------------------------------------------------
// Replacement of EX_patch call implemented as smart ASM macro
// (original code by Extended, modified to become a smart ASM macro by tnussb)
//-----------------------------------------------------------------------------
#undef EX_patch
#define EX_patch(base_addr,tag_ptr) ({\
    asm volatile (\
        "0: move.b -(%1),-1(%%a7)\n"\
        "   move.b -(%1),-(%%a7)\n"\
        "   moveq  #0,%%d0\n"\
        "   move.w (%%a7)+,%%d0\n"\
        "   beq.s  1f\n"\
        "   move.b -(%1),-1(%%a7)\n"\
        "   move.b -(%1),-(%%a7)\n"\
        "   moveq  #0,%%d1\n"\
        "   move.w (%%a7)+,%%d1\n"\
        "   add.l  %0,%%d1\n"\
        "   move.l %%d1,0(%0,%%d0)\n"\
        "   bra.s  0b\n1:" \
        :: "a" ((void*)(base_addr)),"a" ((void*)(tag_ptr)) : "d0","d1");})

#endif

//-----------------------------------------------------------------------------
// embedded version string -- EDITED BY HAND!
//
// NOTE: this version string is embedded in ALL variants and its should !!NOT!!
//       be removed by anyone
//
// to force the compiler to embed the string in all case this variable is
// volatile by intention!
//-----------------------------------------------------------------------------
volatile const char embedded_ttstart_version[]="[ttst121]";

short last_loaded = 0;

//=============================================================================
// handle selection of a demo
//=============================================================================
SYM_ENTRY* HandleDemoSelection() {
    short          previous_font = FontGetSys();
    short          index   = 1;
    short          found   = 0;
    short          i;
    unsigned short length;
    unsigned char* src;
    unsigned char* dest;
    short          popup  = PopupNew((char*)"SELECT a FAT-Demo",C89_92(53,69));
    SYM_ENTRY*     symptr  = SymFindFirst($(main),1);
    char           name[9];
    LCD_BUFFER     screen;

    LCD_save(screen);
    name[8] = 0;

    ClearScreen();
    DrawClipRect(ScrToWin(ScrRect),ScrRect,A_NORMAL);
    DrawStr(C89_92(16,35),3,"FAT-Engine Demo Loader",A_NORMAL);
    FontSetSys(F_4x6);
    DrawStr(C89_92(8,51),C89_92(85,110),"\xA9 2001-2004 TICT (http://tict.ticalc.org)",A_NORMAL);

    while (symptr) {
        if (!strncmp(symptr->name,"fatd",4)) {
            if ((src = HeapDeref(symptr->handle))) {
                length = *(unsigned short*)src+3;
                dest   = src+length-2;

                src += 2;
                if (*dest     == 0xF8 &&
                    *(dest-4) == 'p'  &&
                    *(dest-3) == 'p'  &&
                    *(dest-2) == 'g'  &&
                    ttunpack_valid(src))
                {
                    strncpy(name,symptr->name,8);
                    found++;
                    PopupAddText(popup,-1,name,index);
                }
            }
        }
        symptr=SymFindNext();
        index++;
    }

    if (found<(last_loaded-1)) last_loaded = 0;
    if (found) last_loaded = PopupDo(popup,CENTER,CENTER,last_loaded);
    else {
        DrawStr(C89_92(40,80),30,"No FAT-Demos found !!!",A_REPLACE);
        last_loaded = 0;
        ngetchx();
    }
    ClearScreen();
    HeapFree(popup);
    LCD_restore(screen);
    FontSetSys(previous_font);

    if (!last_loaded) return NULL;

    symptr=SymFindFirst($(main),1);
    for(i=1;i<last_loaded;i++) symptr=SymFindNext();
    return symptr;
}


//=============================================================================
// nothing special, just a huge main routine ...
//=============================================================================
#if __TIGCC_VERSION__ >= 95
__attribute__((section("_st2"))) void main_function(void) {
#else
void _main(void) {
#endif

// #ifndef PARANOID_LAUNCHER
    // This launcher supports the TI-89 Titanium only.
    if ((unsigned long) ROM_base != 0x800000ul) fatal("Wrong calculator model");

//#else
/*
    // This launcher supports the TI-89 Titanium only.
    {
        const HARDWARE_PARM_BLOCK* ptr = FL_getHardwareParmBlock();
        if ((ptr->hardwareID != 9) || (ptr->hardwareRevision != 3)) fatal_without("Wrong calculator model");
    }
*/
// #endif

    unsigned char*         src;
    unsigned char*         dest;
    unsigned short         length;
    SYM_ENTRY*             symptr;
    HANDLE                 h,uncomp_h=0;
    volatile unsigned char mode;// = 0xFF;
#ifndef ONLY_PPG
    unsigned short         asm_prg  = 0;
    unsigned short         do_alloc = 1;
#endif
#if __TIGCC_VERSION__ >= 95
    void*                  program_ptr    = &entry_point - 2;
    HANDLE                 program_handle = HeapPtrToHandle(program_ptr);
#endif

#ifndef REENTRANT
    // let's help the compiler with optimization ...
    short* volatile        its_me_checking_ptr = &its_me_checking;
    // check if "we" are already active (the same copy of this executable) and
    // if yes exit immediately !!
    if (*its_me_checking_ptr) {fatal_without("can't start myself");}
    else                      *its_me_checking_ptr = 1;  // now we are active ...
#endif


    if (!(symptr=HandleDemoSelection())) return;

    //-------------------------------------------------------------------------
    // IMPORTANT: only use handle (symptr may become invalid !!!)
    //-------------------------------------------------------------------------
    h = symptr->handle;

    //-------------------------------------------------------------------------
    // if file is already locked, don't unlock it at the end ...
    //-------------------------------------------------------------------------
    {
        short locked = HeapGetLock(h);
        src = HLock(h);
        if (locked) h = 0;
    }

    if (!src) fatal("locking failed"); // will return!

    length = *(unsigned short*)src+3;
    dest   = src+length-2;

#ifndef ONLY_PPG
    //-------------------------------------------------------------------------
    // check if we want to start a normal ASM program
    //-------------------------------------------------------------------------
    if (*dest == 0xF3) {
        asm_prg = 1;
        dest    = src;
        length--;
        if (!symptr->flags.bits.archived && !symptr->flags.bits.locked) do_alloc = 0;
    }
#endif


#ifdef ONLY_ASM
    else {
        if (h) HeapUnlock(h);
        fatal("no ASM prg"); // will return!
    }

#else

    //-------------------------------------------------------------------------
    // check if given program is not even a packed file
    //-------------------------------------------------------------------------
#ifndef ONLY_PPG
    else
#endif
    {
        //---------------------------------------------------------------------
        // skip "length of variable" bytes
        //---------------------------------------------------------------------
        src += 2;
        if (*dest     != 0xF8 ||
            *(dest-4) != 'p'  ||
            *(dest-3) != 'p'  ||
            *(dest-2) != 'g'  ||
            !ttunpack_valid(src))
        {
            if (h) HeapUnlock(h);
#ifndef ONLY_PPG
            fatal("neither ASM nor PPG prg");
#else
            fatal("no PPG prg");
#endif
        }
        //---------------------------------------------------------------------
        // get the original size and allocate memory for it
        //---------------------------------------------------------------------
        length = ttunpack_size(src);
    }

#endif


#ifndef ONLY_PPG
    if (do_alloc)
#endif
    {
        // NOTE: we don't use malloc, but the low level routines
        //       otherwise compressed kernel-dependent progs will crash
        if (!(uncomp_h=HeapAllocHigh(length+100/*will be rounded up if odd*/))) {
            if (h) HeapUnlock(h);
            fatal("out of memory");
        }

        dest = HeapDeref(uncomp_h);

#ifndef ONLY_PPG
        if (asm_prg) {
            memcpy(dest,src,length);
        }
#endif

#ifndef ONLY_ASM

#ifndef ONLY_PPG
        else
#endif
        {
            ST_helpMsg("decompressing ...");

            if (ttunpack_decompress(src,dest)) {
                //-------------------------------------------------------------
                // "houston we got problems ..."
                //-------------------------------------------------------------
                HeapFree(uncomp_h);
                if (h) HeapUnlock(h);
                fatal("failed");
            }
#ifdef ONLY_PPG
            if (h) HeapUnlock(h);
#endif
            ST_eraseHelp();
        }
    }

#else
    }
#endif

#ifndef TIGCC_PSTARTER
    //-------------------------------------------------------------------------
    // set our own in-use bit, for safety reasons
    //-------------------------------------------------------------------------
#if __TIGCC_VERSION__ >= 95
    if (program_handle) {
        SYM_ENTRY *tmpsym;
        for (tmpsym = SymFindFirst(NULL, FO_RECURSE | FO_SKIP_TEMPS); tmpsym; tmpsym = SymFindNext()) {
            if (tmpsym->handle == program_handle) {
                tmpsym->flags.bits.hidden = 1;
                break;
            }
        }
    }
#endif
#endif

    //-------------------------------------------------------------------------
    // relocate the executable, enter the ghost space and start it ...
    //-------------------------------------------------------------------------
    // Fixed EX_patch call for AMS 2.06+ (thanks Kevin for the fix).
    // Fixed again for HW3/AMS 3.00+ (Kevin).

    EX_patch(dest+2,dest+length-1);

    /* We have to set the "last executed program" to the beginning of the program to
       be executed, or else APD may crash under certain circumstances.
       The code below looks for the "last executed program" variable. That variable
       is cleared during initialization, immediately after the stack fence is set up.
       So we look for the value of the stack fence (0xDEADDEAD) in the initialization
       code and add 8 to get the wanted short pointer, which must then be
       sign-extended to an actual pointer. (The sign extension is implicit in the
       generated code, as it should be.) */
    
    // Lionel: Smaller and faster version of 
/*
    char *p=(char *)0x812000ul;
    while (p<(char *)0x818000ul && *(unsigned long*)p!=0xDEADDEAD) p+=2;

    p+=2[(short *)p]?8:12;
*/

    char *p;
    short i;
    asm("move.w #((0x818000-0x812000)/2)-1,%0 | Set up iteration counter
    lea 0x812000,%1 | Set up pointer
    0:
    cmpi.l #0xDEADDEAD,(%1) | Look for the stack fence
    addq.l #2,%1 | Increment pointer, condition codes unchanged
    dbeq %0,0b
    " : "=d"(i),"=a"(p) : : "cc");

    p+=1[(short *)p]?6:10;
    *(void **)(long)*(short *)p=(void*)((unsigned char *)dest+2);

    unsigned long oldercatcherpc=ercatcherpc;
    TRY
        // The second argument MUST be even.
        ASM_CALL_WITH_RETURN_VALUE_HANDLING(dest+2,(length+1)&0xFFFE,mode);
    FINALLY
        ercatcherpc=oldercatcherpc; /* restore this before rethrowing the error,
                                       in case this is a recursive call */

        // mode must not be changed here !

    //-------------------------------------------------------------------------
    // clear our own in-use bit again; otherwise we are in big trouble
    //-------------------------------------------------------------------------
#if __TIGCC_VERSION__ >= 95
    if (program_handle) {
        SYM_ENTRY *tmpsym;
        for (tmpsym = SymFindFirst(NULL, FO_RECURSE | FO_SKIP_TEMPS); tmpsym; tmpsym = SymFindNext()) {
            if (tmpsym->handle == program_handle) {
                tmpsym->flags.bits.hidden = 0;
                break;
            }
        }
    }
#endif

#ifndef ONLY_PPG
    if (src!=dest)
#endif
        HeapFree(uncomp_h);

#ifndef ONLY_PPG
    if (h) HeapUnlock(h);
#endif



#if __TIGCC_VERSION__ >= 95
    //---------------------------------------------------------------------
    // On AMS 1.xx, we need to unlock our own handle, otherwise we cannot
    // be started again due to an AMS bug.
    //---------------------------------------------------------------------
    if (AMS_1xx) {
        if (program_handle)
            HeapUnlock(program_handle);
    }
#endif

    ENDFINAL

    if (!mode) {
        //---------------------------------------------------------------------
        // launched program uses RETURN_VALUE - handle it correctly
        // (VERY dirty trick implemented by Kevin Kofler)
        // Modified by Lionel Debroux to support ENABLE_ERROR_RETURN and
        // SET_FILE_IN_USE_BIT (the code was broken when these #defines
        // were added).
        // And modified again by Kevin Kofler to use a #define so we don't have
        // to change this everytime the offset changes.
        //---------------------------------------------------------------------
        {
            if (*(((unsigned short **)saved_sp)[SAVED_SP_OFFSET])==0x21ee) ((unsigned short **)saved_sp)[SAVED_SP_OFFSET]+=3;
            else                                                           ((unsigned short **)saved_sp)[SAVED_SP_OFFSET]+=2;
        }
    }
}

//#############################################################################
//###################### NO MORE FAKES BEYOND THIS LINE #######################
//#############################################################################
//
//=============================================================================
// Revision History
//=============================================================================
//
// $Log: fatload.c,v $
// Revision 1.2  2002/05/22 09:41:55  tnussb
// stepped back to old version of unpacking function to fix "random" triggers
// of ERR_Throw
//
// Revision 1.1  2002/05/03 11:32:17  tnussb
// initial version
//
//
//
