/******************************************************************************
*
* project name:    FAT-SDK
* file name:       fatload.c
* initial date:    03/05/2002
* author:          thomas.nussbaumer@gmx.net
* description:     start utility for exepacked FAT-Demos
*
* (based on the code of ttstart from the TIGCC Tools Suite)
*
* $Id: fatload.c,v 1.2 2002/05/22 09:41:55 tnussb Exp $
*
******************************************************************************/

#define USE_TI89       // compile for TI89
#define USE_TI92P      // compile for TI92p
#define USE_V200       // compile for V200

//-----------------------------------------------------------------------------
// Undocumented macro in TIGCC v0.95 and above: Do not create global imports
// for startup code. This macro is defined exclusively for ttstart; do not use
// it in your own programs. In ttstart, it only suppresses many of warnings.
//-----------------------------------------------------------------------------
#define _NO_INCLUDE_PATCH

#define MIN_AMS 100                 // works on all AMS versions
#define NO_EXIT_SUPPORT             // the launcher doesn't need exit support
#define NO_CALC_DETECT              // it doesn't need calc detection, either
#define OPTIMIZE_ROM_CALLS          // generates smaller code by reserving a5
#define ENABLE_ERROR_RETURN         // to return errors to AMS.
#define SET_FILE_IN_USE_BIT         // prevent strange things under some circumstances (see TIGCC help)

#include <version.h>
#include <statline.h>
#include <alloc.h>
#include <vat.h>
#include <string.h>
#include <system.h>
#include <error.h>
#include <estack.h>
#include <compat.h>
#include <menus.h>
#include <graph.h>
#include <kbd.h>

// Starting from TIGCC v0.95, we use our own startup code.
#if __TIGCC_VERSION__ >= 95

// tts121: from void* (would generate warnings with -Wpointer-arith).
unsigned char *real_saved_sp;
#define saved_sp (real_saved_sp-4)
#define SAVED_SP_OFFSET 2 // longwords

asm(".section _st1\n"
"entry_point:"
// OPTIMIZE_ROM_CALLS support.
"	pea.l (%a5)\n"
"	move.l 0xC8,%a5\n"
// Save the stack pointer.
"	lea.l real_saved_sp(%pc),%a0\n"
"	move.l %a7,(%a0)\n"
// Call main_function.
"	bsr.s main_function\n"
// OPTIMIZE_ROM_CALLS support.
"	move.l (%sp)+,%a5\n"
// Return.
"	rts\n"
".text");

extern unsigned char entry_point;

#else

//-----------------------------------------------------------------------------
// DIRTY: This is needed to get everything to work with "complex main"s
//        (Kevin Kofler)
//-----------------------------------------------------------------------------
#if __TIGCCLIB_VERSION__ != 250
#warning Wrong TIGCCLIB version! Do not expect ttstart to work!
#endif
asm("/* Include Patch: save_the_sp */");
extern void *saved_sp asm("__save__sp__");
#ifdef TIGCC_PSTARTER
#define SAVED_SP_OFFSET 18 // longwords
#else
#define SAVED_SP_OFFSET 22 // longwords
#endif

#endif

#define fatal(s) {ST_helpMsg(s); return;}


#define C89_92(x,y) ((TI89)?(x):(y))


typedef struct {
    unsigned char  osize_lo;   // original size lowbyte
    unsigned char  osize_hi;   // original size highbyte
    unsigned char  magic1;     // must be equal to TTUNPACK_MAGIC1
    unsigned char  magic2;     // must be equal to TTUNPACK_MAGIC2
    unsigned char  csize_lo;   // compressed size lowbyte
    unsigned char  csize_hi;   // compressed size lowbyte
    unsigned char  esc1;       // escape >> (8-escBits)
    unsigned char  notused3;
    unsigned char  notused4;
    unsigned char  esc2;       // escBits
    unsigned char  gamma1;     // maxGamma + 1
    unsigned char  gamma2;     // (1<<maxGamma)
    unsigned char  extralz;    // extraLZPosBits
    unsigned char  notused1;
    unsigned char  notused2;
    unsigned char  rleentries; // rleUsed
} TTUNPACK_HEADER;

#define TTUNPACK_MAGIC1 0x54
#define TTUNPACK_MAGIC2 0x50

#define ttunpack_size(_p_)  ((unsigned short)(((TTUNPACK_HEADER*)(_p_))->osize_lo | (((TTUNPACK_HEADER*)(_p_))->osize_hi << 8)))
#define ttunpack_valid(_p_) (((TTUNPACK_HEADER*)(_p_))->magic1 == TTUNPACK_MAGIC1 && ((TTUNPACK_HEADER*)(_p_))->magic2 == TTUNPACK_MAGIC2)

#define TTUNPACK_OKAY             0
#define TTUNPACK_NOESCFOUND     248
#define TTUNPACK_ESCBITS        249
#define TTUNPACK_MAXGAMMA       250
#define TTUNPACK_EXTRALZP       251
#define TTUNPACK_NOMAGIC        252
#define TTUNPACK_OUTBUFOVERRUN  253
#define TTUNPACK_LZPOSUNDERRUN  254

// TODO: upgrade to Sam's assembly routine for a smaller and faster program !
#define ttunpack_decompress ((__attribute__((__stkparm__)) unsigned short(*)(unsigned char*,unsigned char*))(unsigned short[]){\
0x206f,0x0004,0x226f,0x0008,0x48e7,0x010c,0x303c,0x0200,0x4e41,0x3f00,\
0x610e,0x3200,0x301f,0x4e41,0x3001,0x4cdf,0x3080,0x4e75,0x4e56,\
0xfff4,0x48e7,0x1e30,0x2a09,0x0c28,0x0054,0x0002,0x6708,0x303c,0x00fc,\
0x6000,0x021e,0x0c28,0x0050,0x0003,0x66f0,0x0c28,0x0008,0x000a,0x6708,\
0x303c,0x00fa,0x6000,0x0206,0x0c28,0xff80,0x000b,0x66f0,0x4240,0x1028,\
0x0009,0x3d40,0xfff8,0x303c,0x00f9,0x0c6e,0x0008,0xfff8,0x6200,0x01e6,\
0x4240,0x1028,0x000c,0x3d40,0xfff6,0x303c,0x00fb,0x0c6e,0x0004,0xfff6,\
0x6200,0x01ce,0x4240,0x1028,0x0006,0x3640,0x7008,0x906e,0xfff8,0x3d40,\
0xfffa,0x43e8,0x000f,0x2d49,0xfffc,0x4fef,0xfefc,0x2a4f,0x327c,0x0001,\
0x4217,0x4202,0x1002,0x4880,0x7201,0xe169,0x6f0c,0x3001,0x1b82,0x9000,\
0x5249,0x5340,0x66f6,0x5202,0x0c02,0x0007,0x6fe2,0x7000,0x1028,0x000f,\
0x49f0,0x0810,0x3e3c,0x0080,0x4a6e,0xfff8,0x6600,0x0162,0x380b,0xb84b,\
0x6718,0x302e,0xfffa,0xe16c,0x2445,0x5285,0x3200,0x41fa,0x01c6,0x4e90,\
0x8800,0x1484,0x60da,0x426e,0xfff4,0x45fa,0x016a,0x4e92,0x3c00,0x0c40,\
0x0001,0x6600,0x00e2,0x4240,0x1014,0xc047,0x6700,0x00a4,0xe21f,0x6402,\
0x528c,0x4240,0x1014,0xc047,0x676a,0xe21f,0x6402,0x528c,0x4e92,0x3800,\
0x0c40,0x007f,0x6240,0x4e92,0x0c40,0x001f,0x6224,0x0280,0x0000,0xffff,\
0x206e,0xfffc,0x1030,0x0800,0x0240,0x00ff,0x4242,0x2245,0x5285,0x12c0,\
0x5242,0xb842,0x64f4,0x6000,0xff74,0x3600,0x0643,0xffe0,0xe74b,0x7203,\
0x41fa,0x0148,0x4e90,0x8043,0x60da,0x0644,0xff80,0xe34c,0x7201,0x41fa,\
0x0136,0x4e90,0x8840,0x4e92,0x5340,0xe148,0x8840,0x60a6,0xe21f,0x6402,\
0x528c,0x47fa,0x011e,0x322e,0xfff8,0x4e93,0x3600,0x302e,0xfffa,0xe16c,\
0x2445,0x5285,0x3200,0x4e93,0x8800,0x1484,0x3643,0x6000,0xff1c,0xe21f,\
0x6402,0x528c,0x41fa,0x0094,0x4e90,0x3200,0x0a41,0x00ff,0x4242,0x7000,\
0x3001,0x2045,0x91c0,0x102e,0xfff5,0xd028,0xffff,0x2245,0x5285,0x12c0,\
0x5242,0xbc42,0x64e4,0x6000,0xfee8,0x4e92,0x3600,0x5343,0x0c43,0x00fe,\
0x672a,0x4a6e,0xfff6,0x6612,0x41fa,0x0054,0x4e90,0x3200,0x0a41,0x00ff,\
0xe14b,0x8243,0x60ba,0x302e,0xfff6,0xe16b,0x3200,0x41fa,0x009a,0x4e90,\
0x8640,0x60dc,0x0c46,0x0003,0x630e,0x45fa,0x002a,0x4e92,0x3d40,0xfff4,\
0x4e92,0x608e,0x4240,0x6010,0x322e,0xfff8,0x41fa,0x0074,0x4e90,0x3800,\
0x6000,0xfe94,0x4cee,0x0c78,0xffdc,0x4e5e,0x4e75,0x101c,0xe148,0x1014,\
0x1235,0x7000,0x5201,0xe268,0x0240,0x00ff,0x4e75,0x3407,0x1014,0xc400,\
0x673c,0xd402,0x5302,0x1235,0x7000,0xc002,0xb102,0x6608,0x142c,0x0001,\
0x4602,0x5001,0x9235,0x2000,0x1401,0x5202,0x0c01,0x0007,0x6504,0x7207,\
0x7407,0x3007,0xe43f,0xbe00,0x6502,0x528c,0x6100,0x0010,0x5201,0x03c0,\
0x4e75,0x7001,0xe21f,0x6526,0x4e75,0x7000,0x5301,0x6520,0x6732,0x3007,\
0xe23f,0xbe00,0x6518,0xd040,0x5340,0xc01c,0xe148,0x1014,0x1435,0x7800,\
0xe468,0xe21f,0x6402,0x528c,0x4e75,0xd040,0x5340,0xc014,0x1435,0x7800,\
0xe428,0xe21f,0x65ec,0x4e75,0x1014,0xc007,0x56c0,0x0200,0x0001,0xe21f,\
0x65dc,0x4e75})


//-----------------------------------------------------------------------------
// we use an own implementation of enter_ghost_space, because due to the
// arrangements within TIGCCLIB (comes as real compile-time library since 2.4)
// otherwise we cannot generate a compatible pstarter.o anymore
// (the launcher generated with TIGCC option -pack will not be linked against
// tigcc.a)
//-----------------------------------------------------------------------------
void own_enter_ghost_space(void);
asm("own_enter_ghost_space:\n"
"	bsr.w 0f\n"
"0:\n"
"	move.l (%sp)+,%d0\n"
"	cmpi.l #0x40000,%d0\n"
"	bcc.s 2f\n"
"	movea.l 0xC8,%a0\n"
"	cmpi.l #1000,(%a0,-4)\n"
"	bcs.s 2f\n"
"	movem.l %a2-%a6/%d3-%d7,-(%sp)\n"
"	lea (%sp,-20),%sp\n"
"	move.l #0x3E000,%a3\n"
"	move.l %a0,%d0\n"
"	andi.l #0x600000,%d0\n"
"	addi.l #0x20000,%d0\n"
"	move.l %d0,(%sp,12)\n"
"	move.l %d0,(%sp,16)\n"
"	trap #0xC\n"
"	move.w #0x2700,%sr\n"
"	move.l #0xF,%d3\n"
"	pea 1f(%pc)\n"
"	bset.b #2,(%sp,1)\n"
"	clr.w  -(%sp)\n"
"	move.l 0xAC,%a0\n"
"	jmp (%a0)\n"
"1:\n"
"	lea (%sp,20),%sp\n"
"	movem.l (%sp)+,%a2-%a6/%d3-%d7\n"
"2:\n"
"	move.l (%sp)+,%d0\n"
"	bset.l #18,%d0\n"
"	movea.l %d0,%a0\n"
"	jmp (%a0)");


//-----------------------------------------------------------------------------
// The following macro calls function fptr and treats a possible RETURN_VALUE
// correctly. If a return value is available mode is set to 0, otherwise 0xFF.
//
// NOTE: the NOPs are very important due to a dirty hack in the TIGCCLIB with
//       #define RETURN_VALUE (they must be exactly after the call!!)
//
// One nop seems to be enough, that's why I (Lionel Debroux) removed one nop.
//-----------------------------------------------------------------------------

#define ASM_CALL_WITH_RETURN_VALUE_HANDLING(fptr,mode) ({\
        asm volatile ("movem.l %%d1-%%d7/%%a0-%%a6,-(%%sp);"\
            "jsr (%1);"\
            "bra.s 0f;"\
            "nop;"\
            "clr.b %%d0;"\
            "bra.s 1f;"\
            "0: st.b %%d0;"\
            "1: movem.l (%%sp)+,%%d1-%%d7/%%a0-%%a6;"\
            "move.b %%d0,%0"\
            : "=g" (mode) : "a" ((void *)(fptr)) : "d0","cc","memory");})


//-----------------------------------------------------------------------------
// embedded version string -- EDITED BY HAND!
//
// NOTE: this version string is embedded in ALL variants and its should !!NOT!!
//       be removed by anyone
//
// to force the compiler to embed the string in all case this variable is
// volatile by intention!
//-----------------------------------------------------------------------------
volatile const char embedded_ttstart_version[]="[tts121]";

short last_loaded = 0;

//=============================================================================
// handle selection of a demo
//=============================================================================
SYM_ENTRY* HandleDemoSelection(void) {
    short          previous_font = FontGetSys();
    short          index   = 1;
    short          found   = 0;
    short          i;
    unsigned short length;
    unsigned char* src;
    unsigned char* dest;
    HANDLE         popup  = PopupNew((char*)"SELECT a FAT-Demo",C89_92(53,69));
    SYM_ENTRY*     symptr  = SymFindFirst(SYMSTR_CONST("main"),2); // FAT 1.22: 2 from 1
    char           name[9];
    LCD_BUFFER     screen;

    LCD_save(screen);
    name[8] = 0;

    ClearScreen();
    DrawClipRect(ScrToWin(ScrRect),ScrRect,A_NORMAL);
    DrawStr(C89_92(16,35),3,"FAT-Engine Demo Loader",A_NORMAL);
    FontSetSys(F_4x6);
    DrawStr(C89_92(8,51),C89_92(85,110),"\xA9 2001-2004 TICT (http://tict.ticalc.org)",A_NORMAL);

    while (symptr) {
        if (!strncmp(symptr->name,"fatd",4)) {
            if ((src = HeapDeref(symptr->handle))) {
                length = *(unsigned short*)src+3;
                dest   = src+length-2;

                src += 2;
                if (*dest     == 0xF8 &&
                    *(dest-4) == 'p'  &&
                    *(dest-3) == 'p'  &&
                    *(dest-2) == 'g'  &&
                    ttunpack_valid(src))
                {
                    strncpy(name,symptr->name,8);
                    found++;
                    PopupAddText(popup,-1,name,index);
                }
            }
        }
        symptr=SymFindNext();
        index++;
    }

    if (found<(last_loaded-1)) last_loaded = 0;
    if (found) last_loaded = PopupDo(popup,CENTER,CENTER,last_loaded);
    else {
        DrawStr(C89_92(40,80),30,"No FAT-Demos found !!!",A_REPLACE);
        last_loaded = 0;
        ngetchx();
    }
    ClearScreen();
    HeapFree(popup);
    LCD_restore(screen);
    FontSetSys(previous_font);

    if (!last_loaded) return NULL;

    symptr=SymFindFirst(SYMSTR_CONST("main"),1);
    for(i=1;i<last_loaded;i++) symptr=SymFindNext();
    return symptr;
}

//=============================================================================
// nothing special, just a huge main routine ...
//=============================================================================
#if __TIGCC_VERSION__ >= 95
__attribute__((section("_st2"))) void main_function(void) {
#else
void _main(void) {
#endif
    unsigned char*         src;
    unsigned char*         dest;
    unsigned short         length;
    HANDLE                 h,uncomp_h=0;
    volatile unsigned char mode;
#if __TIGCC_VERSION__ >= 95
    void*                  program_ptr    = &entry_point - 2;
    HANDLE                 program_handle = HeapPtrToHandle(program_ptr);
#endif
    SYM_ENTRY*             symptr;

    /* This launcher supports "legacy" TI-89 only. */
    if ((unsigned long) ROM_base == 0x800000ul) fatal("Wrong calculator model");

    if (!(symptr=HandleDemoSelection())) return;

    //-------------------------------------------------------------------------
    // IMPORTANT: only use handle (symptr may become invalid !!!)
    //-------------------------------------------------------------------------
    h = symptr->handle;

    //-------------------------------------------------------------------------
    // if file is already locked, don't lock it again ...
    //-------------------------------------------------------------------------
    if (HeapGetLock(h)) {
        src = HeapDeref(h);
        h   = 0;
    }
    else {
        src = HLock(h);
    }

    if (!src) fatal("locking failed"); // will return!

    length = *(unsigned short*)src+3;
    dest   = src+length-2;


    //---------------------------------------------------------------------
    // skip "length of variable" bytes
    //---------------------------------------------------------------------
    src += 2;
    //---------------------------------------------------------------------
    // get the original size and allocate memory for it
    //---------------------------------------------------------------------
    length = ttunpack_size(src);

    // NOTE: we don't use malloc, but the low level routines
    //       otherwise compressed kernel-dependent progs will crash
    if (!(uncomp_h=HeapAllocHigh(length))) {
        if (h) HeapUnlock(h);
        fatal("out of memory");
    }

    dest = HeapDeref(uncomp_h);

    ST_helpMsg("decompressing ...");

    if (ttunpack_decompress(src,dest)) {
        //-------------------------------------------------------------
        // "houston we got problems ..."
        //-------------------------------------------------------------
        HeapFree(uncomp_h);
        if (h) HeapUnlock(h);
        fatal("failed");
    }
    ST_eraseHelp();

    //-------------------------------------------------------------------------
    // set our own in-use bit, for safety reasons
    //-------------------------------------------------------------------------
#if __TIGCC_VERSION__ >= 95
    if (program_handle) {
        SYM_ENTRY *tmpsym;
        for (tmpsym = SymFindFirst(NULL, FO_RECURSE | FO_SKIP_TEMPS); tmpsym; tmpsym = SymFindNext()) {
            if (tmpsym->handle == program_handle) {
                tmpsym->flags.bits.hidden = 1;
                break;
            }
        }
    }
#endif

    //-------------------------------------------------------------------------
    // relocate the executable, enter the ghostspace and start it ...
    //-------------------------------------------------------------------------
    // FIXED EX_patch call for upcoming AMS 2.06 (thanks Kevin for the fix)

    EX_patch(dest+0x40002,dest+length+0x3FFFF);
    own_enter_ghost_space();
    {(*(unsigned long *)saved_sp)|=0x40000;}

    /* If we are on AMS 2, we have to set the "last executed program" to somewhere
       in the last 4 KB of RAM, or else APD may crash under certain circumstances.
       The code below looks for the "last executed program" variable. That variable
       is cleared during initialization, immediately after the stack fence is set up.
       So we look for the value of the stack fence (0xDEADDEAD) in the initialization
       code and add 8 to get the wanted short pointer, which must then be
       sign-extended to an actual pointer. (The sign extension is implicit in the
       generated code, as it should be.) */
    if (!(AMS_1xx||*(short*)0x32==('R'<<8)+'O')) {
      /* rb and q are factored out in order to get more efficient code. */
      char *rb=ROM_base;
      char *p=rb+0x12000;
      char *q=rb+0x18000;

      while (p<q && *(unsigned long*)p!=0xDEADDEAD) p+=2;
      p+=2[(short *)p]?8:12;
      *(void **)(long)*(short *)p=(void*)0x3f000;
    }

    TRY
        ASM_CALL_WITH_RETURN_VALUE_HANDLING(dest+0x40002,mode);
    FINALLY

    //-------------------------------------------------------------------------
    // clear our own in-use bit again; otherwise we are in big trouble
    //-------------------------------------------------------------------------
#if __TIGCC_VERSION__ >= 95
    if (program_handle) {
        SYM_ENTRY *tmpsym;
        for (tmpsym = SymFindFirst(NULL, FO_RECURSE | FO_SKIP_TEMPS); tmpsym; tmpsym = SymFindNext()) {
            if (tmpsym->handle == program_handle) {
                tmpsym->flags.bits.hidden = 0;
                break;
            }
        }
    }
#endif

    HeapFree(uncomp_h);

    if (h) HeapUnlock(h);

#if __TIGCC_VERSION__ >= 95
    //---------------------------------------------------------------------
    // On AMS 1.xx, we need to unlock our own handle, otherwise we cannot
    // be started again due to an AMS bug.
    //---------------------------------------------------------------------
    if (AMS_1xx) {
        if (program_handle)
            HeapUnlock(program_handle);
    }
#endif

    ENDFINAL

    if (!mode) {
        //---------------------------------------------------------------------
        // launched program uses RETURN_VALUE - handle it correctly
        // (VERY dirty trick implemented by Kevin Kofler)
        // Modified by Lionel Debroux to support ENABLE_ERROR_RETURN and
        // SET_FILE_IN_USE_BIT (the code was broken when these #defines
        // were added).
        // And modified again by Kevin Kofler to use a #define so we don't have
        // to change this everytime the offset changes.
        //---------------------------------------------------------------------
        {
            if (*(((unsigned short **)saved_sp)[SAVED_SP_OFFSET])==0x21ee) ((unsigned short **)saved_sp)[SAVED_SP_OFFSET]+=3;
            else                                                           ((unsigned short **)saved_sp)[SAVED_SP_OFFSET]+=2;
        }
    }
}

//#############################################################################
//###################### NO MORE FAKES BEYOND THIS LINE #######################
//#############################################################################
//
//=============================================================================
// Revision History
//=============================================================================
//
// $Log: fatload.c,v $
// Revision 1.2  2002/05/22 09:41:55  tnussb
// stepped back to old version of unpacking function to fix "random" triggers
// of ERR_Throw
//
// Revision 1.1  2002/05/03 11:32:17  tnussb
// initial version
//
//
//

